##                                                                 -*- Perl -*-
##  2000-2001 Patryk Zadarnowski <pat@jantar.org>
#
# Redistribution  and  use  in  source  and  binary  forms,  with  or  without
# modification, are permitted provided  that the following conditions are met:
# (1) Redistributions of  source code must retain the  above copyright notice,
# this list  of conditions and the following  disclaimer.  (2) Redistributions
# in  binary form  must reproduce  the above  copyright notice,  this  list of
# conditions and  the following disclaimer  in the documentation  and/or other
# materials provided  with the distribution.  (3)  The name of  any author may
# not  be used  to  endorse or  promote  products derived  from this  software
# without their specific prior written permission.
#
# THIS SOFTWARE IS  PROVIDED ``AS IS'' AND ANY  EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
# AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
# SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
# INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
# CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
# ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# This perl script parses the "ArabicShaping.txt" file and generates the
# following tables that map out the Arabic shaping classes of characters:
#
#	const __unidata_iso646_t __unidata_arabic_shaping_pool[];
#	const __unidata_soff_t   __unidata_arabic_schematic_name[65536];
#	const __unidata_prop_t   __unidata_arabic_joining_type[65536];
#	const __unidata_soff_t   __unidata_arabic_joining_group[65536];
#
# The joining type property is encoded as follows:
#
#	0	R	right-joining
#	1	L	left-joining
#	2	D	dual-joining
#	3	C	join-causing
#	4	U	non-joining
#	5	T	transparent
#
# The data in "ArabicShaping.txt" seems incomplete and does not include all of
# the information from 8 of The Unicode Standard, Version 3.0.  The above
# tables do not attempt to fill in the missing data, and map all unlisted
# characters to the joinging type 'U' and the goining group "<no shaping>".
# I ident to complement it with the information from the various tables in
# 8 of The Unicode Standard in the future.

my $progress_limit;
my $progress_current;
my $progress_running = 0;

sub begin_progress_bar {
    $progress_current = "  0";
    $progress_running = 1;
    $progress_limit = shift;
    print STDERR "@_: $progress_current\%";
}

sub update_progress_bar {
    my $x = sprintf "\%3.0f", ($_[0] / $progress_limit * 100);
    if ($x ne $progress_current) {
	printf STDERR "\b\b\b\b$x\%";
	$progress_current = $x;
    }
}

sub end_progress_bar {
    print "\b\b\b\b100\%\n";
    $progress_running = 0;
}

sub panic {
    my $header = shift;
    print STDERR "\n" if $progress_running;
    print STDERR "[ArabicShaping.pl] $header \n";
    foreach $line (@_) {
	print STDERR "\t$line\n";
    }
    exit 1;
}

sub addslashes {
    $_[0] =~ s/[\"\\]/\\\&/g;
    return $_[0];
}

# Read and parse the input file:

my $src = "ArabicShaping";

begin_progress_bar 65536, "Loading \"$src\"";
open IN, "< $src.txt" or panic ("Cannot open \"$src.txt\":", $!);

my %unidata_arabic_shaping_pool = ();
my @unidata_arabic_schematic_name = ();
my @unidata_arabic_joining_type = ();
my @unidata_arabic_joining_group = ();

my %joining_type_encoding = (
    "R" => 0,
    "L" => 1,
    "D" => 2,
    "C" => 3,
    "U" => 4,
    "T" => 5
);

my $seq = 0;
my $lineno = 0;

while (<IN>) {

    ++$lineno;

    # Remove comments, and trailing spaces:
    s/\s*(\#.*)?$//gs;
    next if !$_;

    # Parse the line:
    my @line = split /\s*;\s*/;
    my $code = hex $line[0];
    my $schematic_name = addslashes $line[1];
    my $joining_type   = $line[2];
    my $joining_group  = addslashes $line[3];

    # Skip characters ouside of the BMP:
    next if $code > 0xFFFF;

    # Handle unallocated characters:
    while ($seq < $code) {
	$unidata_arabic_schematic_name[$seq] = "";
	$unidata_arabic_joining_type[$seq] = "U";
	$unidata_arabic_joining_group[$seq] = "<no shaping>";
	$unidata_arabic_shaping_pool{""} = -1;
	$unidata_arabic_shaping_pool{"<no shaping>"} = -1;
	update_progress_bar ++$seq;
    }

    # Handle this character:
    $unidata_arabic_schematic_name[$seq] = $schematic_name;
    $unidata_arabic_joining_type[$seq] = $joining_type;
    $unidata_arabic_joining_group[$seq] = $joining_group;
    $unidata_arabic_shaping_pool{$schematic_name} = -1;
    $unidata_arabic_shaping_pool{$joining_group} = -1;
    update_progress_bar ++$seq;
}
while ($seq <= 0xFFFF) {
    $unidata_arabic_schematic_name[$seq] = "";
    $unidata_arabic_joining_type[$seq] = "U";
    $unidata_arabic_joining_group[$seq] = "<no shaping>";
    $unidata_arabic_shaping_pool{""} = -1;
    $unidata_arabic_shaping_pool{"<no shaping>"} = -1;
    update_progress_bar ++$seq;
}
end_progress_bar;

# Generate the C file:
print STDERR "Generating \"$src.c\":\n";
open OUT, "> $src.c" or panic "Cannot create \"$src.c\"", $!;
print OUT "/* This file has been automatically generated. Do not edit! */

#include \"unidata.h\"

";

# Generate the string pool:
$table = "arabic_shaping_pool";
@list = sort keys %unidata_arabic_shaping_pool;
begin_progress_bar scalar(@list), "  Generating \"__unidata_$table\"";
print OUT "const __unidata_iso646_t __unidata_${table}[] =";
$offset = 0;
$i = 0;
foreach $string (@list) {
    $unidata_arabic_shaping_pool{$string} = $offset;
    print OUT "\n    \"$string\\0\"";
    update_progress_bar (++$i);
    $offset += length($string) + 1;
}
print OUT ";

";
end_progress_bar;

# Generate the "__unidata_arabic_schematic_name" table:
$table = "arabic_schematic_name";
# Generate the schematic name table:
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_soff_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 8 == 0;
    printf OUT "%6d, ",
        $unidata_arabic_shaping_pool{$unidata_arabic_schematic_name[$i]};
    update_progress_bar $i;
}
print OUT "
};

";
end_progress_bar;

# Generate the "__unidata_arabic_joining_type" table:
$table = "arabic_joining_type";
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_prop_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 16 == 0;
    print OUT "$joining_type_encoding{$unidata_arabic_joining_type[$i]}, ";
    update_progress_bar $i;
}
print OUT "
};

";
end_progress_bar;

# Generate the "__unidata_arabic_joining_group" table:
$table = "arabic_joining_group";
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_soff_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 8 == 0;
    printf OUT "%6d, ",
        $unidata_arabic_shaping_pool{$unidata_arabic_joining_group[$i]};
    update_progress_bar $i;
}
print OUT "
};

";
end_progress_bar;

print STDERR "  String pool size: $offset bytes\n";
close(OUT);
