##                                                                 -*- Perl -*-
##  2000-2001 Patryk Zadarnowski <pat@jantar.org>
#
# Redistribution  and  use  in  source  and  binary  forms,  with  or  without
# modification, are permitted provided  that the following conditions are met:
# (1) Redistributions of  source code must retain the  above copyright notice,
# this list  of conditions and the following  disclaimer.  (2) Redistributions
# in  binary form  must reproduce  the above  copyright notice,  this  list of
# conditions and  the following disclaimer  in the documentation  and/or other
# materials provided  with the distribution.  (3)  The name of  any author may
# not  be used  to  endorse or  promote  products derived  from this  software
# without their specific prior written permission.
#
# THIS SOFTWARE IS  PROVIDED ``AS IS'' AND ANY  EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
# AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
# SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
# INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
# CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
# ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# This perl script parses the "BidiMirroring.txt" file and generates the
# following tables that map out the bidirectional mirroring properties of
# characters:
#
#	const __unidata_prop_t  __unidata_bidi_mapping_type[65536];
#	const __unidata_utf16_t __unidata_bidi_mirror[65536];
#
# The mapping type is encoded as follows:
#
#	0	characters without the "mirrored" property
#	1	mirrored characters without a mapping within the BMP
#	2	mirrored characters with a best-fit mapping
#	3	mirrored characters with a proper mapping
#
# The "mirrored" property comes from the tenth field in "UnicodeData.txt".
# The remaining data comes from "BidiMirroring.txt", suplemented by mapping
# all characters without a mirrored image to themselves.

my $progress_limit;
my $progress_current;
my $progress_running = 0;

sub begin_progress_bar {
    $progress_current = "  0";
    $progress_running = 1;
    $progress_limit = shift;
    print STDERR "@_: $progress_current\%";
}

sub update_progress_bar {
    my $x = sprintf "\%3.0f", ($_[0] / $progress_limit * 100);
    if ($x ne $progress_current) {
	printf STDERR "\b\b\b\b$x\%";
	$progress_current = $x;
    }
}

sub end_progress_bar {
    print "\b\b\b\b100\%\n";
    $progress_running = 0;
}

sub panic {
    my $header = shift;
    print STDERR "\n" if $progress_running;
    print STDERR "[BidiMirroring.pl] $header \n";
    foreach $line (@_) {
	print STDERR "\t$line\n";
    }
    exit 1;
}

sub addslashes {
    $_[0] =~ s/[\"\\]/\\\&/g;
    return $_[0];
}

# Read and parse "UnicodeData.txt":
my $src = "UnicodeData";
begin_progress_bar 65536, "Loading \"$src\"";
open IN, "< $src.txt" or panic ("Cannot open \"$src.txt\":", $!);

my @unidata_bidi_mapping_type = ();
my @unidata_bidi_mirror = ();

my $seq = 0;
my $lineno = 0;
my $in_range = 0;

while (<IN>) {

    ++$lineno;

    # Remove comments, and trailing spaces:
    s/\s*(\#.*)?$//gs;
    next if !$_;

    # Parse the line:
    my @line = split /\s*;\s*/;
    my $code = hex $line[0];
    my $mirrored = ($line[9] eq "Y") ? 1 : 0;

    # Skip characters ouside of the BMP:
    next if $code > 0xFFFF;

    # Handle unallocated characters:
    if (!$in_range) {
	while ($seq < $code) {
	    $unidata_bidi_mapping_type[$seq] = 0;
	    $unidata_bidi_mirror[$seq] = $seq;
	    update_progress_bar ++$seq;
	}
    }

    # Handle this character or character range:
    if ($name =~ /, First>$/) {
	# Handle a character range:
	!$in_range or panic ("Error on line $lineno of \"UnicodeData.txt\":",
			     "Unterminated code range.");
	$in_range = 1;
    }
    elsif ($name =~ /, Last>$/) {
	# Complete the range:
	$in_range or panic ("Error on line $lineno of \"UnicodeData.txt\":",
			    "Unexpected end of code range.");
	while ($seq <= $code) {
	    $unidata_bidi_mapping_type[$seq] = $mirrored;
	    $unidata_bidi_mirror[$seq] = $seq;
	    update_progress_bar ++$seq;
	}
	$in_range = 0;
    }
    else {
	# Handle a single character:
	!$in_range or panic ("Error on line $lineno of \"UnicodeData.txt\":",
			     "Unterminated code range.");
	$unidata_bidi_mapping_type[$seq] = $mirrored;
	$unidata_bidi_mirror[$seq] = $seq;
	update_progress_bar ++$seq;
    }
}
while ($seq <= 0xFFFF) {
    $unidata_bidi_mapping_type[$seq] = 0;
    $unidata_bidi_mirror[$seq] = $seq;
    update_progress_bar ++$seq;
}
end_progress_bar;
close IN;

# Patch the collected information using data from "BidiMirroring.txt":
my $src = "BidiMirroring";
begin_progress_bar 65536, "Loading \"$src\"";
open IN, "< $src.txt" or panic ("Cannot open \"$src.txt\":", $!);

my $seq = 0;
my $lineno = 0;

while (<IN>) {

    ++$lineno;

    # Extract the mapping kind from the comment field:
    $bidi_mapping_type = /\[BEST FIT\]/ ? 2 : 3;

    # Remove comments, and trailing spaces:
    s/\s*(\#.*)?$//gs;
    next if !$_;

    # Parse the line:
    my @line = split /\s*;\s*/;
    my $code = hex $line[0];
    my $bidi_mirror = hex $line[1];

    # Skip characters ouside of the BMP:
    next if $code > 0xFFFF || $mirror > 0xFFFF;

    # Handle this character: (skip unallocated characters)
    $unidata_bidi_mapping_type[$code] = $bidi_mapping_type;
    $unidata_bidi_mirror[$code] = $bidi_mirror;
    update_progress_bar $code;
}
end_progress_bar;

# Generate the C file:
print STDERR "Generating \"$src.c\":\n";
open OUT, "> $src.c" or panic "Cannot create \"$src.c\"", $!;
print OUT "/* This file has been automatically generated. Do not edit! */

#include \"unidata.h\"

";

# Generate the "__unidata_bidi_mapping_type" table:
$table = "bidi_mapping_type";
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_prop_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 16 == 0;
    print OUT "$unidata_bidi_mapping_type[$i], ";
    update_progress_bar $i;
}
print OUT "
};

";
end_progress_bar;

# Generate the "__unidata_bidi_mirror" table:
$table = "bidi_mirror";
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_utf16_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 8 == 0;
    printf OUT "0x%04X, ", $unidata_bidi_mirror[$i];
    update_progress_bar $i;
}
print OUT "
};

";
end_progress_bar;

close(OUT);
