##                                                                 -*- Perl -*-
##  2000-2001 Patryk Zadarnowski <pat@jantar.org>
#
# Redistribution  and  use  in  source  and  binary  forms,  with  or  without
# modification, are permitted provided  that the following conditions are met:
# (1) Redistributions of  source code must retain the  above copyright notice,
# this list  of conditions and the following  disclaimer.  (2) Redistributions
# in  binary form  must reproduce  the above  copyright notice,  this  list of
# conditions and  the following disclaimer  in the documentation  and/or other
# materials provided  with the distribution.  (3)  The name of  any author may
# not  be used  to  endorse or  promote  products derived  from this  software
# without their specific prior written permission.
#
# THIS SOFTWARE IS  PROVIDED ``AS IS'' AND ANY  EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
# AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
# SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
# INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
# CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
# ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# This perl script parses the "Blocks.txt" file and generates the list of
# block names, 
#
#	const int                __unidata_block_count;
#	const __unidata_iso646_t __unidata_block_pool[];
#	const __unidata_utf16_t  __unidata_block_posn[];
#	const __unidata_soff_t   __unidata_block_name[];
#
#
# This information is structured in a somewhat unusual fashion. The BMP is
# divided into non-overlapping blocks, with bogus ``reserved'' blocks inserted
# to ensure a complete cover by the partitioning. The number of blocks,
# including the reserved ones, is stored in "__unidata_block_count". The first
# code position of each block is stored in "__unidata_block_posn".  To make it
# possible to compute the block size, an additional U+0000 entry has been
# appended to "__unidata_block_posn", which can be used both as a list
# terminator and to compute the block size as a difference of the two
# consecutive positions, modulo 65536. The name of each block is stored in
# "__unidata_block_name", as a byte offset into "__unidata_block_pool".
#
# Unallocated blocks are given names consisting of an empty string.

my $progress_limit;
my $progress_current;
my $progress_running = 0;

sub begin_progress_bar {
    $progress_current = "  0";
    $progress_running = 1;
    $progress_limit = shift;
    print STDERR "@_: $progress_current\%";
}

sub update_progress_bar {
    my $x = sprintf "\%3.0f", ($_[0] / $progress_limit * 100);
    if ($x ne $progress_current) {
	printf STDERR "\b\b\b\b$x\%";
	$progress_current = $x;
    }
}

sub end_progress_bar {
    print "\b\b\b\b100\%\n";
    $progress_running = 0;
}

sub panic {
    my $header = shift;
    print STDERR "\n" if $progress_running;
    print STDERR "[Blocks.pl] $header \n";
    foreach $line (@_) {
	print STDERR "\t$line\n";
    }
    exit 1;
}

sub addslashes {
    $_[0] =~ s/[\"\\]/\\\&/g;
    return $_[0];
}

# Read and parse the "Blocks.txt" file:

my $src = "Blocks";

begin_progress_bar 65536, "Loading \"$src.txt\"";
open IN, "< $src.txt" or panic ("Cannot open \"src.txt\":", $!);

my $unidata_block_count = 0;
my %unidata_block_pool = ();
my @unidata_block_posn = ();
my @unidata_block_name = ();

my $seq = 0;
my $lineno = 0;

while (<IN>) {

    ++$lineno;

    # Remove comments, and trailing spaces:
    s/\s*(\#.*)?$//gs;
    next if !$_;

    # Parse the line:
    my @line = split /\s*;\s*/;
    my $start = hex $line[0];
    my $end   = hex $line[1];
    my $name  = addslashes $line[2];

    # Skip characters ouside of the BMP:
    next if $start > 0xFFFF;
    $end = 0xFFFF if $end > 0xFFFF;

    # Handle unallocated characters:
    if ($seq != $start) {
	$unidata_block_posn[scalar @unidata_block_posn] = $seq;
	$unidata_block_name[scalar @unidata_block_name] = "";
	$unidata_block_pool{""} = -1;
	++$unidata_block_count;
	update_progress_bar $start;
    }

    # Handle this character or character range:
    $unidata_block_posn[scalar @unidata_block_posn] = $start;
    $unidata_block_name[scalar @unidata_block_name] = $name;
    $unidata_block_pool{$name} = -1;
    ++$unidata_block_count;

    $seq = $end + 1;
    update_progress_bar $seq;
}
if ($seq <= 0xFFFF) {
    $unidata_block_posn[scalar @unidata_block_posn] = $seq;
    $unidata_block_name[scalar @unidata_block_name] = "";
    $unidata_block_pool{""} = -1;
    ++$unidata_block_count;
}
$unidata_block_posn[scalar @unidata_block_posn] = 0;
end_progress_bar;

# Generate "blocks.c":
print STDERR "Generating \"$src.c\":\n";
open OUT, "> $src.c" or panic "Cannot create \"$src.c\"", $!;
print OUT "/* This file has been automatically generated. Do not edit! */

#include \"unidata.h\"

";

begin_progress_bar 1, "  Generating \"__unidata_block_count\"";
print OUT "const int __unidata_block_count = $unidata_block_count;

";
end_progress_bar;

# Generate the string pool:
$table = "block_pool";
@list = sort keys %unidata_block_pool;
begin_progress_bar scalar(@list), "  Generating \"__unidata_$table\"";
print OUT "const __unidata_iso646_t __unidata_${table}[] =";
$offset = 0;
$i = 0;
foreach $string (@list) {
    $unidata_block_pool{$string} = $offset;
    print OUT "\n    \"$string\\0\"";
    update_progress_bar (++$i);
    $offset += length($string) + 1;
}
print OUT ";

";
end_progress_bar;

# Generate the name table:
$table = "block_name";
begin_progress_bar $unidata_block_count, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_soff_t __unidata_${table}[] = {";
for ($i = 0; $i < $unidata_block_count; ++$i) {
    print OUT "\n    " if $i % 8 == 0;
    printf OUT "%6d, ", $unidata_block_pool{$unidata_block_name[$i]};
    update_progress_bar $i;
}
print OUT "
};

";
end_progress_bar;

# Generate the position table:
$table = "block_posn";
begin_progress_bar ($unidata_block_count + 1,
		    "  Generating \"__unidata_$table\"");
print OUT "const __unidata_utf16_t __unidata_${table}[] = {";
for ($i = 0; $i < $unidata_block_count + 1; ++$i) {
    print OUT "\n    " if $i % 8 == 0;
    printf OUT "0x%04X, ", $unidata_block_posn[$i];
    update_progress_bar $i;
}
print OUT "
};

";
end_progress_bar;

print STDERR "  Number of blocks: $unidata_block_count\n";
print STDERR "  String pool size: $offset bytes\n";
close(OUT);
