##                                                                 -*- Perl -*-
##  2000-2001 Patryk Zadarnowski <pat@jantar.org>
#
# Redistribution  and  use  in  source  and  binary  forms,  with  or  without
# modification, are permitted provided  that the following conditions are met:
# (1) Redistributions of  source code must retain the  above copyright notice,
# this list  of conditions and the following  disclaimer.  (2) Redistributions
# in  binary form  must reproduce  the above  copyright notice,  this  list of
# conditions and  the following disclaimer  in the documentation  and/or other
# materials provided  with the distribution.  (3)  The name of  any author may
# not  be used  to  endorse or  promote  products derived  from this  software
# without their specific prior written permission.
#
# THIS SOFTWARE IS  PROVIDED ``AS IS'' AND ANY  EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
# AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
# SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
# INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
# CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
# ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# This perl script parses the "CaseFolding.txt" file and generates the
# following tables:
#
#	const __unidata_utf16_t __unidata_neutral_case_pool[];
#	const __unidata_prop_t  __unidata_neutral_case_mapping_type[65536];
#	const __unidata_loff_t  __unidata_neutral_case[65536];
#
# The mapping type is encoded as follows:
#
#	0	no case mapping
#	1	lower-case mapping
#	2	exceptional mapping
#
# The mapping defined in "CaseFolding.txt" maps each code position to a
# sequence of characters, representing the lowercase (or, in some cases,
# exceptional) mapping of the character, eliminating all case differences
# between strings as described in UTR #21.

my $progress_limit;
my $progress_current;
my $progress_running = 0;

sub begin_progress_bar {
    $progress_current = "  0";
    $progress_running = 1;
    $progress_limit = shift;
    print STDERR "@_: $progress_current\%";
}

sub update_progress_bar {
    my $x = sprintf "\%3.0f", ($_[0] / $progress_limit * 100);
    if ($x ne $progress_current) {
	printf STDERR "\b\b\b\b$x\%";
	$progress_current = $x;
    }
}

sub end_progress_bar {
    print "\b\b\b\b100\%\n";
    $progress_running = 0;
}

sub panic {
    my $header = shift;
    print STDERR "\n" if $progress_running;
    print STDERR "[CaseFolding.pl] $header \n";
    foreach $line (@_) {
	print STDERR "\t$line\n";
    }
    exit 1;
}

sub addslashes {
    $_[0] =~ s/[\"\\]/\\\&/g;
    return $_[0];
}

# Read and parse the input file:

my $src = "CaseFolding";

begin_progress_bar 65536, "Loading \"$src.txt\"";
open IN, "< $src.txt" or panic ("Cannot open \"$src.txt\":", $!);

my %unidata_neutral_case_pool = ();
my @unidata_neutral_case_mapping_type = ();
my @unidata_neutral_case = ();

my $seq = 0;
my $lineno = 0;

while (<IN>) {

    ++$lineno;

    # Remove comments, and trailing spaces:
    s/\s*(\#.*)?$//gs;
    next if !$_;

    # Parse the line:
    my @line = split /\s*;\s*/;
    my $code = hex $line[0];
    my $mapping_type = ($line[1] eq "L") ? 1 : 2;
    my $mapping = $line[2];

    # Process unlisted characters:
    while ($seq < $code) {
	my $x = sprintf "0x%04X", $seq;
	$unidata_neutral_case_pool{$x} = -1;
	$unidata_neutral_case_mapping_type[$seq] = 0;
	$unidata_neutral_case[$seq] = $x;
	update_progress_bar ++$seq;
    }

    # Process this character:
    @mappings = split /\s+/, $mapping;
    if (scalar(@mappings) == 0) {
	$char = sprintf "%04X", $code;
	panic "Error on line $lineno of \"src.txt\":",
	    "Null case mapping of U+$char";
    }
    my $x = "0x" . join ", 0x", @mappings;
    $unidata_neutral_case_pool{$x} = -1;
    $unidata_neutral_case_mapping_type[$seq] = $mapping_type;
    $unidata_neutral_case[$seq] = $x;
    update_progress_bar ++$seq;
}
while ($seq <= 0xFFFF) {
    my $x = sprintf "0x%04X", $seq;
    $unidata_neutral_case_pool{$x} = -1;
    $unidata_neutral_case_mapping_type[$seq] = 0;
    $unidata_neutral_case[$seq] = $x;
    update_progress_bar ++$seq;
}
end_progress_bar;

# Generate "case.c":
print STDERR "Generating \"$src.c\":\n";
open OUT, "> $src.c" or panic "Cannot create \"$src.c\"", $!;
print OUT "/* This file has been automatically generated. Do not edit! */

#include \"unidata.h\"

";

# Generate the string pool:
$table = "neutral_case_pool";
@list = sort keys %unidata_neutral_case_pool;
begin_progress_bar scalar(@list), "  Generating \"__unidata_$table\"";
print OUT "const __unidata_utf16_t __unidata_${table}[] = {\n";
$offset = 0;
$i = 0;
foreach $string (@list) {
    $unidata_neutral_case_pool{$string} = $offset;
    print OUT "    $string, 0xFFFF,\n";
    update_progress_bar (++$i);
    $offset += (scalar(split(/\s+/, $string)) + 1) * 2;
}
print OUT "};

";
end_progress_bar;

# Generate the __unidata_neutral_case_mapping_type table:
$table = "neutral_case_mapping_type";
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_prop_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 16 == 0;
    print OUT "$unidata_neutral_case_mapping_type[$i], ";
    update_progress_bar $i;
}
print OUT "
};

";
end_progress_bar;

# Generate the __unidata_neutral_case table:
$table = "neutral_case";
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_loff_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 8 == 0;
    printf OUT "%6d, ", $unidata_neutral_case_pool{$unidata_neutral_case[$i]};
    update_progress_bar $i;
}
print OUT "
};

";
end_progress_bar;

print STDERR "  String pool size: $offset bytes\n";
close(OUT);
