##                                                                 -*- Perl -*-
##  2000-2001 Patryk Zadarnowski <pat@jantar.org>
#
# Redistribution  and  use  in  source  and  binary  forms,  with  or  without
# modification, are permitted provided  that the following conditions are met:
# (1) Redistributions of  source code must retain the  above copyright notice,
# this list  of conditions and the following  disclaimer.  (2) Redistributions
# in  binary form  must reproduce  the above  copyright notice,  this  list of
# conditions and  the following disclaimer  in the documentation  and/or other
# materials provided  with the distribution.  (3)  The name of  any author may
# not  be used  to  endorse or  promote  products derived  from this  software
# without their specific prior written permission.
#
# THIS SOFTWARE IS  PROVIDED ``AS IS'' AND ANY  EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
# AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
# SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
# INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
# CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
# ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# This perl script parses the "CompositionExclusions.txt" file and generates the
# following tables:
#
#	const __unidata_prop_t __unidata_composition_exclusion[65536];
#
# The composition exclusion property is encoded as follows:
#
#	0	not excluded
#	1	script-specific exclusion
#	2	post-composition version character
#	3	singleton decomposition
#	4	non-starter decomposition
#
# Properties (1) and (2) are extracted from "CompositionExclusions.txt".  The
# remaining properties are extracted from "UnicodeData.txt" as described in
# "CompositionExclusions.txt".

my $progress_limit;
my $progress_current;
my $progress_running = 0;

sub begin_progress_bar {
    $progress_current = "  0";
    $progress_running = 1;
    $progress_limit = shift;
    print STDERR "@_: $progress_current\%";
}

sub update_progress_bar {
    my $x = sprintf "\%3.0f", ($_[0] / $progress_limit * 100);
    if ($x ne $progress_current) {
	printf STDERR "\b\b\b\b$x\%";
	$progress_current = $x;
    }
}

sub end_progress_bar {
    print "\b\b\b\b100\%\n";
    $progress_running = 0;
}

sub panic {
    my $header = shift;
    print STDERR "\n" if $progress_running;
    print STDERR "[CompositionExclusions.pl] $header \n";
    foreach $line (@_) {
	print STDERR "\t$line\n";
    }
    exit 1;
}

sub addslashes {
    $_[0] =~ s/[\"\\]/\\\&/g;
    return $_[0];
}

# Read and parse "UnicodeData.txt":
my $src = "UnicodeData";
begin_progress_bar 65536, "Loading \"$src\"";
open IN, "< $src.txt" or panic ("Cannot open \"$src.txt\":", $!);

my @unidata_composition_exclusion = ();

# Non-starter decompositions are computed from the first character of the
# decomposition mapping and the combining class of that character.  We load
# that data into the following two arrays, and patch it into the composition
# exclusion table later.

my @unidata_decomp_mapping = ();
my @unidata_combclass = ();

my $seq = 0;
my $lineno = 0;
my $in_range = 0;

while (<IN>) {

    ++$lineno;

    # Remove comments, and trailing spaces:
    s/\s*(\#.*)?$//gs;
    next if !$_;

    # Parse the line:
    my @line = split /\s*;\s*/;
    my $code = hex $line[0];
    my $combclass = int $line[3];
    my $decomp = $line[5];

    # Skip characters ouside of the BMP:
    next if $code > 0xFFFF;

    # Handle unallocated characters:
    if (!$in_range) {
	while ($seq < $code) {
	    $unidata_composition_exclusion[$seq] = 0;
	    $unidata_decomp_mapping[$seq] = 0;
	    $unidata_combclass[$seq] = 0;
	    update_progress_bar (++$seq);
	}
    }

    # Compute the exclusion type:
    if ($decomp eq "" || $decomp =~ /^<([^>]+)>(.*)/) {
	# No canonical decomposition:
	$type = 0;
	$mapping = 0;
    }
    else {
	@mapping = map { hex $_ } (split / /, $decomp);
	if (scalar(@mapping) > 1) {
	    # Possible non-starter decomposition
	    $type = 4;
	    $mapping = @mapping[0];
	}
	else {
	    # Singleton decomposition:
	    $type = 3;
	    $mapping = 0;
	}
    }

    # Handle this character or character range:
    if ($name =~ /, First>$/) {
	# Handle a character range:
	!$in_range or panic ("Error on line $lineno of \"UnicodeData.txt\":",
			     "Unterminated code range.");
	$in_range = 1;
    }
    elsif ($name =~ /, Last>$/) {
	# Complete the range:
	$in_range or panic ("Error on line $lineno of \"UnicodeData.txt\":",
			    "Unexpected end of code range.");
	while ($seq <= $code) {
	    $unidata_composition_exclusion[$seq] = $type;
	    $unidata_decomp_mapping[$seq] = $mapping;
	    $unidata_combclass[$seq] = $combclass;
	    update_progress_bar (++$seq);
	}
	$in_range = 0;
    }
    else {
	# Handle a single character:
	!$in_range or panic ("Error on line $lineno of \"UnicodeData.txt\":",
			     "Unterminated code range.");
	$unidata_composition_exclusion[$seq] = $type;
	$unidata_decomp_mapping[$seq] = $mapping;
	$unidata_combclass[$seq] = $combclass;
	update_progress_bar (++$seq);
    }
}
while ($seq <= 0xFFFF) {
    $unidata_composition_exclusion[$seq] = 0;
    $unidata_decomp_mapping[$seq] = 0;
    $unidata_combclass[$seq] = 0;
    update_progress_bar (++$seq);
}
end_progress_bar;
close IN;

# Patch the non-starter exclusions:
for ($seq = 0; $seq < 65536; ++$seq) {
    if ($unidata_composition_exclusion[$seq] == 4) {
	if ($unidata_combclass[$unidata_decomp_mapping[$seq]] <= 0) {
	    $unidata_composition_exclusion[$seq] = 0;
	}
    }
}

# Patch the collected information using data from "CompositionExclusions.txt":
my $src = "CompositionExclusions";
begin_progress_bar 65536, "Loading \"$src\"";
open IN, "< $src.txt" or panic ("Cannot open \"$src.txt\":", $!);

my $seq = 0;
my $lineno = 0;

my $type = 1;

while (<IN>) {

    ++$lineno;

    # Switch to mapping type 2 after a "magic" comment:
    $type = 2 if /\# \(2\) Post Composition Version characters/;

    # Remove comments, and trailing spaces:
    s/\s*(\#.*)?$//gs;
    next if !$_;

    # Parse the line:
    my $code = hex;

    # Skip characters ouside of the BMP:
    next if $code > 0xFFFF || $mirror > 0xFFFF;

    # Handle this character: (skip unallocated characters)
    $unidata_composition_exclusion[$code] = $type;
    update_progress_bar $code;
}
end_progress_bar;

# Generate the C file:
print STDERR "Generating \"$src.c\":\n";
open OUT, "> $src.c" or panic "Cannot create \"$src.c\"", $!;
print OUT "/* This file has been automatically generated. Do not edit! */

#include \"unidata.h\"

";

# Generate the "__unidata_composition_exclusion" table:
$table = "composition_exclusion";
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_prop_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 16 == 0;
    print OUT "$unidata_composition_exclusion[$i], ";
    update_progress_bar $i;
}
print OUT "
};

";
end_progress_bar;

close(OUT);
