##                                                                 -*- Perl -*-
##  2000-2001 Patryk Zadarnowski <pat@jantar.org>
#
# Redistribution  and  use  in  source  and  binary  forms,  with  or  without
# modification, are permitted provided  that the following conditions are met:
# (1) Redistributions of  source code must retain the  above copyright notice,
# this list  of conditions and the following  disclaimer.  (2) Redistributions
# in  binary form  must reproduce  the above  copyright notice,  this  list of
# conditions and  the following disclaimer  in the documentation  and/or other
# materials provided  with the distribution.  (3)  The name of  any author may
# not  be used  to  endorse or  promote  products derived  from this  software
# without their specific prior written permission.
#
# THIS SOFTWARE IS  PROVIDED ``AS IS'' AND ANY  EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
# AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
# SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
# INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
# CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
# ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# This perl script parses the "EastAsianWidth.txt" file and generates the
# following tables that map out the East Asian width property of characters:
#
#	const __unidata_prop_t   __unidata_east_asian_width[65536];
#
# The property is encoded as follows:
#
#	0	N	East Asian neutral
#	1	A	East Asian ambiguous
#	2	H	East Asian half-width
#	3	W	East Asian wide
#	4	F	East Asian full-width
#	5	Na	East Asian narrow
#
# All ulisted characters are mapped to 'N'.

my $progress_limit;
my $progress_current;
my $progress_running = 0;

sub begin_progress_bar {
    $progress_current = "  0";
    $progress_running = 1;
    $progress_limit = shift;
    print STDERR "@_: $progress_current\%";
}

sub update_progress_bar {
    my $x = sprintf "\%3.0f", ($_[0] / $progress_limit * 100);
    if ($x ne $progress_current) {
	printf STDERR "\b\b\b\b$x\%";
	$progress_current = $x;
    }
}

sub end_progress_bar {
    print "\b\b\b\b100\%\n";
    $progress_running = 0;
}

sub panic {
    my $header = shift;
    print STDERR "\n" if $progress_running;
    print STDERR "[EastAsianWidth.pl] $header \n";
    foreach $line (@_) {
	print STDERR "\t$line\n";
    }
    exit 1;
}

sub addslashes {
    $_[0] =~ s/[\"\\]/\\\&/g;
    return $_[0];
}

# Read and parse the input file:

my $src = "EastAsianWidth";

begin_progress_bar 65536, "Loading \"$src\"";
open IN, "< $src.txt" or panic ("Cannot open \"$src.txt\":", $!);

my @unidata_east_asian_width = ();

my %width_encoding = (
    "N"  => 0, # East Asian neutral
    "A"  => 1, # East Asian ambiguous
    "H"  => 2, # East Asian half-width
    "W"  => 3, # East Asian wide
    "F"  => 4, # East Asian full-width
    "Na" => 5  # East Asian narrow
);

my $seq = 0;
my $lineno = 0;
my $in_range = 0;

while (<IN>) {

    ++$lineno;

    # Remove comments, and trailing spaces:
    s/\s*(\#.*)?$//gs;
    next if !$_;

    # Parse the line:
    my @line = split /\s*;\s*/;
    my $code = hex $line[0];
    my $width = $width_encoding{$line[1]};
    my $name = $line[2];

    # Skip characters ouside of the BMP:
    next if $code > 0xFFFF;

    # Handle unallocated characters:
    if (!$in_range) {
	while ($seq < $code) {
	    $unidata_east_asian_width[$seq] = 0;
	    update_progress_bar ++$seq;
	}
    }

    # Handle this character or character range:
    if ($name =~ /, First>$/) {
	# Handle a character range:
	!$in_range or panic("Error on line $lineno of \"EastAsianWidth.txt\":",
			    "Unterminated code range.");
	$in_range = 1;
    }
    elsif ($name =~ /, Last>$/) {
	# Complete the range:
	$in_range or panic("Error on line $lineno of \"EastAsianWidth.txt\":",
			   "Unexpected end of code range.");
	while ($seq <= $code) {
	    $unidata_east_asian_width[$seq] = $width;
	    update_progress_bar ++$seq;
	}
	$in_range = 0;
    }
    else {
	# Handle a single character:
	!$in_range or panic("Error on line $lineno of \"EastAsianWidth.txt\":",
			    "Unterminated code range.");
	$unidata_east_asian_width[$seq] = $width;
	update_progress_bar ++$seq;
    }
}
while ($seq <= 0xFFFF) {
    $unidata_east_asian_width[$seq] = 0;
    update_progress_bar ++$seq;
}
end_progress_bar;

# Generate the C file:
print STDERR "Generating \"$src.c\":\n";
open OUT, "> $src.c" or panic "Cannot create \"$src.c\"", $!;
print OUT "/* This file has been automatically generated. Do not edit! */

#include \"unidata.h\"

";

# Generate the "__unidata_east_asian_width" table:
$table = "east_asian_width";
# Generate the schematic name table:
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_prop_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 16 == 0;
    print OUT "$unidata_east_asian_width[$i], ";
    update_progress_bar $i;
}
print OUT "
};

";
end_progress_bar;

close(OUT);
