##                                                                 -*- Perl -*-
##  2000-2001 Patryk Zadarnowski <pat@jantar.org>
#
# Redistribution  and  use  in  source  and  binary  forms,  with  or  without
# modification, are permitted provided  that the following conditions are met:
# (1) Redistributions of  source code must retain the  above copyright notice,
# this list  of conditions and the following  disclaimer.  (2) Redistributions
# in  binary form  must reproduce  the above  copyright notice,  this  list of
# conditions and  the following disclaimer  in the documentation  and/or other
# materials provided  with the distribution.  (3)  The name of  any author may
# not  be used  to  endorse or  promote  products derived  from this  software
# without their specific prior written permission.
#
# THIS SOFTWARE IS  PROVIDED ``AS IS'' AND ANY  EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
# AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
# SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
# INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
# CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
# ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# This perl script parses the "Jamo.txt" file and generates the following
# tables that map out the Jamo short name property.
#
#	const __unidata_iso646_t __unidata_jamo_pool[];
#	const __unidata_soff_t   __unidata_jamo_short_name[256];
#
# Only the Jamo block is mapped out. Unallocated characters are mapped to an
# empty string.

my $progress_limit;
my $progress_current;
my $progress_running = 0;

sub begin_progress_bar {
    $progress_current = "  0";
    $progress_running = 1;
    $progress_limit = shift;
    print STDERR "@_: $progress_current\%";
}

sub update_progress_bar {
    my $x = sprintf "\%3.0f", ($_[0] / $progress_limit * 100);
    if ($x ne $progress_current) {
	printf STDERR "\b\b\b\b$x\%";
	$progress_current = $x;
    }
}

sub end_progress_bar {
    print "\b\b\b\b100\%\n";
    $progress_running = 0;
}

sub panic {
    my $header = shift;
    print STDERR "\n" if $progress_running;
    print STDERR "[Jamo.pl] $header \n";
    foreach $line (@_) {
	print STDERR "\t$line\n";
    }
    exit 1;
}

sub addslashes {
    $_[0] =~ s/[\"\\]/\\\&/g;
    return $_[0];
}

# Read and parse the "Blocks.txt" file:

my $src = "Jamo";

begin_progress_bar 256, "Loading \"$src.txt\"";
open IN, "< $src.txt" or panic ("Cannot open \"$src.txt\":", $!);

$start = 0x1100;	# the first character in the Hangul Jamo block
$end   = 0x11FF;	# the last character in the Hangul Jamo block

my %unidata_jamo_pool = ();
my @unidata_jamo_short_name = ();

my $seq = $start;
my $lineno = 0;

while (<IN>) {

    ++$lineno;

    # Remove comments, and trailing spaces:
    s/\s*(\#.*)?$//gs;
    next if !$_;

    # Parse the line:
    my @line = split /\s*;\s*/;
    my $code = hex $line[0];
    my $name = addslashes $line[1];

    # Skip characters ouside of the Hangul Jamo block:
    next if $code > $end;

    # Handle unallocated characters:
    while ($seq < $code) {
	$unidata_jamo_pool{""} = -1;
	$unidata_jamo_short_name[$seq - $start] = "";
	update_progress_bar (++$seq - $start);
    }

    # Handle this character:
    $unidata_jamo_pool{$name} = -1;
    $unidata_jamo_short_name[$seq - $start] = $name;
    update_progress_bar (++$seq - $start);
}
while ($seq <= $end) {
    $unidata_jamo_pool{""} = -1;
    $unidata_jamo_short_name[$seq - $start] = "";
    update_progress_bar (++$seq - $start);
}
end_progress_bar;

# Generate "jamo.c":
print STDERR "Generating \"$src.c\":\n";
open OUT, "> $src.c" or panic "Cannot create \"$src.c\"", $!;
print OUT "/* This file has been automatically generated. Do not edit! */

#include \"unidata.h\"

";

# Generate the string pool:
$table = "jamo_pool";
@list = sort keys %unidata_jamo_pool;
begin_progress_bar scalar(@list), "  Generating \"__unidata_$table\"";
print OUT "const __unidata_iso646_t __unidata_jamo_${table}[] =";
$offset = 0;
$i = 0;
foreach $string (@list) {
    $unidata_jamo_pool{$string} = $offset;
    print OUT "\n    \"$string\\0\"";
    update_progress_bar (++$i);
    $offset += length($string) + 1;
}
print OUT ";

";
end_progress_bar;

# Generate the name table:
$table = "jamo_short_name";
begin_progress_bar 256, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_soff_t __unidata_${table}[] = {";
for ($i = 0; $i < 256; ++$i) {
    print OUT "\n    " if $i % 8 == 0;
    printf OUT "%6d, ", $unidata_jamo_pool{$unidata_jamo_short_name[$i]};
    update_progress_bar $i;
}
print OUT "
};

";
end_progress_bar;

print STDERR "  String pool size: $offset bytes\n";
close(OUT);
