##                                                                 -*- Perl -*-
##  2000-2001 Patryk Zadarnowski <pat@jantar.org>
#
# Redistribution  and  use  in  source  and  binary  forms,  with  or  without
# modification, are permitted provided  that the following conditions are met:
# (1) Redistributions of  source code must retain the  above copyright notice,
# this list  of conditions and the following  disclaimer.  (2) Redistributions
# in  binary form  must reproduce  the above  copyright notice,  this  list of
# conditions and  the following disclaimer  in the documentation  and/or other
# materials provided  with the distribution.  (3)  The name of  any author may
# not  be used  to  endorse or  promote  products derived  from this  software
# without their specific prior written permission.
#
# THIS SOFTWARE IS  PROVIDED ``AS IS'' AND ANY  EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
# AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
# SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
# INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
# CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
# ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# This perl script parses the "LineBreak.txt" file and generates the
# following tables that map out the line break properties of characters:
#
#	const __unidata_prop_t  __unidata_line_break_type[65536];
#
# The property is encoded as follows:
#
#	0	XX	unknown
#	1	BK	mandatory break
#	2	CR	carriage return
#	3	LF	line feed
#	4	CM	attached characters and combining marks  marks
#	5	SG	surrogates
#	6	ZW	zero width space
#	7	IN	inseparable
#	8	GL	non-breaking (glue)
#	9	CB	contingent break opportunity
#	10	SP	space
#	11	BA	break opportunity after
#	12	BB	break opportunity before
#	13	B2	break opportunity before and after
#	14	HY	hyphen
#	15	NS	non starter
#	16	OP	opening punctuation
#	17	CL	closing punctuation
#	18	QU	ambiguous quotation
#	19	EX	exclamation/interrogation
#	20	ID	ideographic
#	21	NU	numeric
#	22	IS	infix separator (numeric)
#	23	SY	symbols allowing breaks
#	24	AL	ordinary alphabetic and symbol characters
#	25	PR	prefix (numeric)
#	26	PO	postfix (numeric)
#	27	SA	complex context (south east asian)
#	28	AI	ambiguous (alphabetic or ideographic)
#
# Assigned characters that are not listed explicitly are given the value AL.
# Unassigned characters are given the value "XX".  The list of unassigned
# characters is obtained from "UnicodeData.txt".

my $progress_limit;
my $progress_current;
my $progress_running = 0;

sub begin_progress_bar {
    $progress_current = "  0";
    $progress_running = 1;
    $progress_limit = shift;
    print STDERR "@_: $progress_current\%";
}

sub update_progress_bar {
    my $x = sprintf "\%3.0f", ($_[0] / $progress_limit * 100);
    if ($x ne $progress_current) {
	printf STDERR "\b\b\b\b$x\%";
	$progress_current = $x;
    }
}

sub end_progress_bar {
    print "\b\b\b\b100\%\n";
    $progress_running = 0;
}

sub panic {
    my $header = shift;
    print STDERR "\n" if $progress_running;
    print STDERR "[LineBreak.pl] $header \n";
    foreach $line (@_) {
	print STDERR "\t$line\n";
    }
    exit 1;
}

sub addslashes {
    $_[0] =~ s/[\"\\]/\\\&/g;
    return $_[0];
}

# Read and parse "UnicodeData.txt":
my $src = "UnicodeData";
begin_progress_bar 65536, "Loading \"$src\"";
open IN, "< $src.txt" or panic ("Cannot open \"$src.txt\":", $!);

my @unidata_line_break_type = ();

my $seq = 0;
my $lineno = 0;
my $in_range = 0;

%type_encoding = (
    "XX" =>  0, # unknown
    "BK" =>  1, # mandatory break
    "CR" =>  2, # carriage return
    "LF" =>  3, # line feed
    "CM" =>  4, # attached characters and combining marks  marks
    "SG" =>  5, # surrogates
    "ZW" =>  6, # zero width space
    "IN" =>  7, # inseparable
    "GL" =>  8, # non-breaking (glue)
    "CB" =>  9, # contingent break opportunity
    "SP" => 10, # space
    "BA" => 11, # break opportunity after
    "BB" => 12, # break opportunity before
    "B2" => 13, # break opportunity before and after
    "HY" => 14, # hyphen
    "NS" => 15, # non starter
    "OP" => 16, # opening punctuation
    "CL" => 17, # closing punctuation
    "QU" => 18, # ambiguous quotation
    "EX" => 19, # exclamation/interrogation
    "ID" => 20, # ideographic
    "NU" => 21, # numeric
    "IS" => 22, # infix separator (numeric)
    "SY" => 23, # symbols allowing breaks
    "AL" => 24, # ordinary alphabetic and symbol characters
    "PR" => 25, # prefix (numeric)
    "PO" => 26, # postfix (numeric)
    "SA" => 27, # complex context (south east asian)
    "AI" => 28  # ambiguous (alphabetic or ideographic)
);

my $type = $type_encoding{"AL"};

while (<IN>) {

    ++$lineno;

    # Remove comments, and trailing spaces:
    s/\s*(\#.*)?$//gs;
    next if !$_;

    # Parse the line:
    my @line = split /\s*;\s*/;
    my $code = hex $line[0];

    # Skip characters ouside of the BMP:
    next if $code > 0xFFFF;

    # Handle unallocated characters:
    if (!$in_range) {
	while ($seq < $code) {
	    $unidata_line_break_type[$seq] = 0;
	    update_progress_bar ++$seq;
	}
    }

    # Handle this character or character range:
    if ($name =~ /, First>$/) {
	# Handle a character range:
	!$in_range or panic ("Error on line $lineno of \"UnicodeData.txt\":",
			     "Unterminated code range.");
	$in_range = 1;
    }
    elsif ($name =~ /, Last>$/) {
	# Complete the range:
	$in_range or panic ("Error on line $lineno of \"UnicodeData.txt\":",
			    "Unexpected end of code range.");
	while ($seq <= $code) {
	    $unidata_line_break_type[$seq] = $type;
	    update_progress_bar ++$seq;
	}
	$in_range = 0;
    }
    else {
	# Handle a single character:
	!$in_range or panic ("Error on line $lineno of \"UnicodeData.txt\":",
			     "Unterminated code range.");
	$unidata_line_break_type[$seq] = $type;
	update_progress_bar ++$seq;
    }
}
while ($seq <= 0xFFFF) {
    $unidata_line_break_type[$seq] = 0;
    update_progress_bar ++$seq;
}
end_progress_bar;
close IN;

# Patch the collected information using data from "LineBreak.txt":
my $src = "LineBreak";
begin_progress_bar 65536, "Loading \"$src\"";
open IN, "< $src.txt" or panic ("Cannot open \"$src.txt\":", $!);

my $seq = 0;
my $lineno = 0;
my $in_range = 0;

while (<IN>) {

    ++$lineno;

    # Remove comments, and trailing spaces:
    s/\s*(\#.*)?$//gs;
    next if !$_;

    # Parse the line:
    my @line = split /\s*;\s*/;
    my $code = hex $line[0];
    my $type = $type_encoding{$line[1]};

    # Skip characters ouside of the BMP:
    next if $code > 0xFFFF;

    # Skip unlisted characters:
    if (!$in_range) {
	$seq = $code;
    }

    # Handle this character or character range:
    if ($name =~ /, First>$/) {
	# Handle a character range:
	!$in_range or panic ("Error on line $lineno of \"LineBreak.txt\":",
			     "Unterminated code range.");
	$in_range = 1;
    }
    elsif ($name =~ /, Last>$/) {
	# Complete the range:
	$in_range or panic ("Error on line $lineno of \"LineBreak.txt\":",
			    "Unexpected end of code range.");
	while ($seq <= $code) {
	    $unidata_line_break_type[$seq] = $type;
	    update_progress_bar ++$seq;
	}
	$in_range = 0;
    }
    else {
	# Handle a single character:
	!$in_range or panic ("Error on line $lineno of \"LineBreak.txt\":",
			     "Unterminated code range.");
	$unidata_line_break_type[$seq] = $type;
	update_progress_bar ++$seq;
    }
}
end_progress_bar;

# Generate the C file:
print STDERR "Generating \"$src.c\":\n";
open OUT, "> $src.c" or panic "Cannot create \"$src.c\"", $!;
print OUT "/* This file has been automatically generated. Do not edit! */

#include \"unidata.h\"

";

# Generate the "__unidata_line_break_type" table:
$table = "line_break_type";
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_prop_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 16 == 0;
    printf OUT "%2d, ", $unidata_line_break_type[$i];
    update_progress_bar $i;
}
print OUT "
};

";
end_progress_bar;

close(OUT);
