##                                                                 -*- Perl -*-
##  2000-2001 Patryk Zadarnowski <pat@jantar.org>
#
# Redistribution  and  use  in  source  and  binary  forms,  with  or  without
# modification, are permitted provided  that the following conditions are met:
# (1) Redistributions of  source code must retain the  above copyright notice,
# this list  of conditions and the following  disclaimer.  (2) Redistributions
# in  binary form  must reproduce  the above  copyright notice,  this  list of
# conditions and  the following disclaimer  in the documentation  and/or other
# materials provided  with the distribution.  (3)  The name of  any author may
# not  be used  to  endorse or  promote  products derived  from this  software
# without their specific prior written permission.
#
# THIS SOFTWARE IS  PROVIDED ``AS IS'' AND ANY  EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
# AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
# SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
# INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
# CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
# ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# This perl script parses the "UnicodeData.txt" file and generates the
# following information:
#
# 1. CHARACTER NAMES
# ==================
#
#	const __unidata_iso646_t __unidata_name_pool[];
#	const __unidata_loff_t   __unidata_name[65536];
#
# The data is extracted from the second field of "UnicodeData.txt",
# amended with the following special cases:
#
# 1. All undefined code positions are named "<reserved>".
# 2. The two code positions U+FFFE and U+FFFF are named "<not a character">.
# 3. The names of Hangul Syllables can be derived algorithmically,
#    and are named "<Hangul syllable>" in the name table.
# 3. Similarly, CJK Unified Ideaographs are named "<CJK ideograph>"
# 4. It does not make sense to name surrogates and private-use
#    code positions, and therefore these code positions are named
#    "<surrogate>" and "<private>" respectively.
#
# I do not attempt to compress the name data, although avoid duplication of
# the pseudo-names enclosed in angle brackets, as they are assigned to ranges
# of characters.
#
# 2. GENERAL CATEGORIES
# =====================
#
#	const __unidata_prop_t __unidata_general_category[65536];
#
# The property is encoded as follows:
#
#	0	Lu	Letter, Uppercase
#	1	Ll	Letter, Lowercase
#	2	Lt	Letter, Titlecase
#	3	Mn	Mark, Non-Spacing
#	4	Mc	Mark, Spacing Combining
#	5	Me	Mark, Enclosing
#	6	Nd	Number, Decimal Digit
#	7	Nl	Number, Letter
#	8	No	Number, Other
#	9	Zs	Separator, Space
#	10	Zl	Separator, Line
#	11	Zp	Separator, Paragraph
#	12	Cc	Other, Control
#	13	Cf	Other, Format
#	14	Cs	Other, Surrogate
#	15	Co	Other, Private Use
#	16	Cn	Other, Not Assigned
#	17	Lm	Letter, Modifier
#	18	Lo	Letter, Other
#	19	Pc	Punctuation, Connector
#	20	Pd	Punctuation, Dash
#	21	Ps	Punctuation, Open
#	22	Pe	Punctuation, Close
#	23	Pi	Punctuation, Initial quote
#	24	Pf	Punctuation, Final quote
#	25	Po	Punctuation, Other
#	26	Sm	Symbol, Math
#	27	Sc	Symbol, Currency
#	28	Sk	Symbol, Modifier
#	29	So	Symbol, Other
#
# The data is extracted from the third field of "UnicodeData.txt".  All
# undefined code positions, as well as the two non-characters U+FFFE and
# U+FFFF, are assigned to the category Cn.
#
# 3. CANONICAL COMBINING CLASSES
# ==============================
#
#	const __unidata_prop_t __unidata_combining_class[65536];
#
# The numeric encoding of this property is defined by Unicode Consorium.  The
# data is extracted from the fourth field of "UnicodeData.txt".  All undefined
# code positions, U+FFFE and U+FFFF are assigned to the combining class 0.
#
# 4. BIDIRECTIONAL CATEGORY
# =========================
#
#	const __unidata_prop_t __unidata_bidi_category[65536];
#
# The property is encoded as follows:
#
#	0	L	Left-to-Right
#	1	LRE	Left-to-Right Embedding
#	2	LRO	Left-to-Right Override
#	3	R	Right-to-Left
#	4	AL	Right-to-Left Arabic
#	5	RLE	Right-to-Left Embedding
#	6	RLO	Right-to-Left Override
#	7	PDF	Pop Directional Format
#	8	EN	European Number
#	9	ES	European Number Separator
#	10	ET	European Number Terminator
#	11	AN	Arabic Number
#	12	CS	Common Number Separator
#	13	NSM	Non-Spacing Mark
#	14	BN	Boundary Neutral
#	15	B	Paragraph Separator
#	16	S	Segment Separator
#	17	WS	Whitespace
#	18	ON	Other Neutrals
#
# The data is extracted from the fifth field of "UnicodeData.txt".  All
# undefined code positions, U+FFFE and U+FFFF are assigned to the
# bidirectional category L (Left-to-Right)
#
# 5. CHARACTER DECOMPOSITION MAPPING
# ==================================
#
#	const __unidata_utf16_t __unidata_decomp_pool[];
#	const __unidata_prop_t  __unidata_decomp_type[65536];
#	const __unidata_loff_t  __unidata_decomp[65536];
#
# The property is encoded as follows:
#
#	0	none	 No decomposition mapping
#	1	canon	 A canononical decomposition mapping
#	2	font	 A font variant (eg. a blackletter form)
#	3	noBreak	 A no-break version of a space or hyphen
#	4	initial	 An initial presentation form (Arabic)
#	5	medial	 A medial presentation form (Arabic)
#	6	final	 A final presentation form (Arabic)
#	7	isolated An isolated presentation form (Arabic)
#	8	circle	 An encircled form
#	9	super	 A superscript form
#	10	sub	 A subscript form
#	11	vertical A vertical layout presentation form
#	12	wide	 A wide (or zenkaku) compatibility character
#	13	narrow	 A narrow (or hankaku) compatibility character
#	14	small	 A small variant form (CNS compatibility)
#	15	square	 A CJK squared font variant
#	16	fraction A vulgar fraction form
#	17	compat	 Otherwise unspecified compatibility character
#
# The data is extracted from the sixth field of "UnicodeData.txt".
# All undefined code positions, U+FFFE and U+FFFF are not decomposed.
#
# 6. DECIMAL DIGIT VALUE
# ======================
#
#	const __unidata_digit_t __unidata_decimal_value[65536];
#
# The data is extracted from the seventh field of "UnicodeData.txt". All
# characters without a decimal digit value are mapped to the NaN 255.
#
# 7. DIGIT VALUE
# ==============
#
#	const __unidata_digit_t __unidata_digit_value[65536];
#
# The data is extracted from the eighth field of "UnicodeData.txt". All
# characters without a digit value are mapped to the NaN 255.
#
# 8. NUMERIC VALUE
# ================
#
#	const __unidata_nat_t __unidata_numeric_value[65536*2];
#
# Each rational value is stored as a pair of natural numbers, with the
# numerator for character x stored in "__unidata_numeric[2*x]" and the
# denominator stored in "__unidata_numeric[2*z+1]".
#
# The data is extracted from the ninth field of "UnicodeData.txt". All
# All characters without a digit value are mapped to the NaN 0/0.
#
# 9. UPPERCASE MAPPING
# ====================
#
#	const __unidata_utf16_t __unidata_upper_case[65536];
#
# The data is extracted from the twelfth field of "UnicodeData.txt".
# All characters without an uppercase mapping are mapped to themselves.
#
# 10. LOWERCASE MAPPING
# =====================
#
#	const __unidata_utf16_t __unidata_lower_case[65536];
#
# The data is extracted from the thirteenth field of "UnicodeData.txt".
# All characters without an lowercase mapping are mapped to themselves.
#
# 11. TITLECASE MAPPING
# =====================
#
#	const __unidata_utf16_t __unidata_title_case[65536];
#
# The data is extracted from the fourteenth field of "UnicodeData.txt".
# All characters without an titlecase mapping are mapped to themselves.
#
# The "mirrored" property is processed by "BidiMirroring.pl". The informative
# "ISO 10646 Comment" and "Unicode 1.0 Name" properties are currently
# unmapped, and will remain so until someone requests their inclusion in the
# library.

my $progress_limit;
my $progress_current;
my $progress_running = 0;

sub begin_progress_bar {
    $progress_current = "  0";
    $progress_running = 1;
    $progress_limit = shift;
    print STDERR "@_: $progress_current\%";
}

sub update_progress_bar {
    my $x = sprintf "\%3.0f", ($_[0] / $progress_limit * 100);
    if ($x ne $progress_current) {
	printf STDERR "\b\b\b\b$x\%";
	$progress_current = $x;
    }
}

sub end_progress_bar {
    print "\b\b\b\b100\%\n";
    $progress_running = 0;
}

sub panic {
    my $header = shift;
    print STDERR "\n" if $progress_running;
    print STDERR "[UnicodeData.pl] $header \n";
    foreach $line (@_) {
	print STDERR "\t$line\n";
    }
    exit 1;
}

sub addslashes {
    $_[0] =~ s/[\"\\]/\\\&/g;
    return $_[0];
}

# Read and parse the "UnicodeData.txt" file:

my $src = "UnicodeData";

begin_progress_bar 65536, "Loading \"$src.txt\"";
open IN, "< $src.txt" or panic ("Cannot open \"$src.txt\":", $!);

my %unidata_name_pool = ();		# the name pool
my %unidata_decomp_pool = ();		# the decomposition pool
my @unidata_name = ();			# the name map
my @unicode_general_category = ();	# the general category map
my @unicode_combingin_class = ();	# the combining class map
my @unicode_bidi_category = ();		# the bidirectional category map
my @unidata_decomp_type = ();		# the decomposition map
my @unidata_decomp = ();		# the decomposition map
my @unicode_decimal_value = ();		# the decimal digit value map
my @unicode_digit_value = ();		# the digit value map
my @unicode_numeric_value = ();		# the numeric value map
my @unicode_upper_case = ();		# the upper-case mapping
my @unicode_lower_case = ();		# the lower-case mapping
my @unicode_title_case = ();		# the title-case mapping

%gencat_encoding = (
    "Lu" => 0,		# Letter, Uppercase
    "Ll" => 1,		# Letter, Lowercase
    "Lt" => 2,		# Letter, Titlecase
    "Mn" => 3,		# Mark, Non-Spacing
    "Mc" => 4,		# Mark, Spacing Combining
    "Me" => 5,		# Mark, Enclosing
    "Nd" => 6,		# Number, Decimal Digit
    "Nl" => 7,		# Number, Letter
    "No" => 8,		# Number, Other
    "Zs" => 9,		# Separator, Space
    "Zl" => 10,		# Separator, Line
    "Zp" => 11,		# Separator, Paragraph
    "Cc" => 12,		# Other, Control
    "Cf" => 13,		# Other, Format
    "Cs" => 14,		# Other, Surrogate
    "Co" => 15,		# Other, Private Use
    "Cn" => 16,		# Other, Not Assigned
    "Lm" => 17,		# Letter, Modifier
    "Lo" => 18,		# Letter, Other
    "Pc" => 19,		# Punctuation, Connector
    "Pd" => 20,		# Punctuation, Dash
    "Ps" => 21,		# Punctuation, Open
    "Pe" => 22,		# Punctuation, Close
    "Pi" => 23,		# Punctuation, Initial quote
    "Pf" => 24,		# Punctuation, Final quote
    "Po" => 25,		# Punctuation, Other
    "Sm" => 26,		# Symbol, Math
    "Sc" => 27,		# Symbol, Currency
    "Sk" => 28,		# Symbol, Modifier
    "So" => 29		# Symbol, Other
);

%bidicat_encoding = (
    "L" => 0,		# Left-to-Right
    "LRE" => 1,		# Left-to-Right Embedding
    "LRO" => 2,		# Left-to-Right Override
    "R" => 3,		# Right-to-Left
    "AL" => 4,		# Right-to-Left Arabic
    "RLE" => 5,		# Right-to-Left Embedding
    "RLO" => 6,		# Right-to-Left Override
    "PDF" => 7,		# Pop Directional Format
    "EN" => 8,		# European Number
    "ES" => 9,		# European Number Separator
    "ET" => 10,		# European Number Terminator
    "AN" => 11,		# Arabic Number
    "CS" => 12,		# Common Number Separator
    "NSM" => 13,		# Non-Spacing Mark
    "BN" => 14,		# Boundary Neutral
    "B" => 15,		# Paragraph Separator
    "S" => 16,		# Segment Separator
    "WS" => 17,		# Whitespace
    "ON" => 18		# Other Neutrals
);

%decomp_encoding = (
    "none" => 0,	# No decomposition mapping
    "canon" => 1,	# A canononical decomposition mapping
    "font" => 2,	# A font variant (eg. a blackletter form)
    "noBreak" => 3,	# A no-break version of a space or hyphen
    "initial" => 4,	# An initial presentation form (Arabic)
    "medial" => 5,	# A medial presentation form (Arabic)
    "final" => 6,	# A final presentation form (Arabic)
    "isolated" => 7,	# An isolated presentation form (Arabic)
    "circle" => 8,	# An encircled form
    "super" => 9,	# A superscript form
    "sub" => 10,	# A subscript form
    "vertical" => 11,	# A vertical layout presentation form
    "wide" => 12,	# A wide (or zenkaku) compatibility character
    "narrow" => 13,	# A narrow (or hankaku) compatibility character
    "small" => 14,	# A small variant form (CNS compatibility)
    "square" => 15,	# A CJK squared font variant
    "fraction" => 16,	# A vulgar fraction form
    "compat" => 17	# Otherwise unspecified compatibility character
);

my $seq = 0; 		# the current character
my $in_range = 0;	# true while processing a character range
my $lineno = 0;		# current line number

while (<IN>) {

    ++$lineno;

    # Remove comments, and trailing spaces:
    s/\s*(\#.*)?$//gs;
    next if !$_;

    # Parse the line:
    my @line = split /\s*;\s*/;
    my $code = hex $line[0];
    my $name = addslashes $line[1];
    my $gencat = $line[2];
    my $combclass = int $line[3];
    my $bidicat = $line[4];
    my $decomp = $line[5];
    my $decimal = ($line[6] eq "") ? 255 : int $line[6];
    my $digit = ($line[7] eq "") ? 255 : int $line[7];
    my $numeric = $line[8];
    my $upper = ($line[11] eq "") ? $code : hex $line[11];
    my $lower = ($line[12] eq "") ? $code : hex $line[12];
    my $title = ($line[13] eq "") ? $code : hex $line[13];

    # Skip characters ouside of the BMP:
    next if $code > 0xFFFF;

    # Handle unallocated characters:
    if (!$in_range) {
	while ($seq < $code) {
	    $unidata_name_pool{"<reserved>"} = -1;
	    $unidata_decomp_pool{""} = -1;
	    $unidata_name[$seq] = "<reserved>";
	    $unidata_general_category[$seq] = $gencat_encoding{"Cn"};
	    $unidata_combining_class[$seq] = 0;
	    $unidata_bidi_category[$seq] = $bidicat_encoding{"L"};
	    $unidata_decomp_type[$seq] = $decomp_encoding{"none"};
	    $unidata_decomp[$seq] = "";
	    $unidata_decimal_value[$seq] = 255;
	    $unidata_digit_value[$seq] = 255;
	    $unidata_numeric_value[2*$seq] = 0;
	    $unidata_numeric_value[2*$seq+1] = 0;
	    $unidata_upper_case[$seq] = $seq;
	    $unidata_lower_case[$seq] = $seq;
	    $unidata_title_case[$seq] = $seq;
	    update_progress_bar ++$seq;
	}
    }

    # Parse the decomposition mapping:
    if ($decomp eq "") {
	# No decomposition:
	$decomp_type = "none";
	$decomp = "";
    }
    else {
	if ($decomp =~ /^<([^>]+)>(.*)/) {
	    # Compatibility decomposition:
	    $decomp_type = $1;
	    $decomp = $2;
	    $decomp =~ s/^\s+//;
	    $decomp or panic ("Error on line $lineno of \"$src.txt\":",
			      "Empty compatiblity decomposition.");
	}
	else {
	    # Canonical decomposition:
	    $decomp_type = "canon";
	}
	# Convert to a list of hexadecimal literals:
	@list = split / /, $decomp;
	$decomp = "0x" . (join ", 0x", @list) . ", ";
    }

    # Parse the numeric value:
    if ($numeric eq "") {
	$numer = 0;
	$denom = 0;
    }
    elsif ($numeric =~ /([0-9]+)\/([0-9]+)/) {
	$numer = int $1;
	$denom = int $2;
    }
    else {
	$numer = int $numeric;
	$denom = 1;
    }

    # Handle this character or character range:
    if ($name =~ /, First>$/) {
	# Handle a character range:
	!$in_range or panic ("Error on line $lineno of \"$src.txt\":",
			     "Unterminated code range.");
	$in_range = 1;
    }
    elsif ($name =~ /, Last>$/) {
	# Complete the range:
	$in_range or panic ("Error on line $lineno of \"$src.txt\":",
			    "Unexpected end of code range.");
	if ($name =~ /^<CJK Ideograph/) {
	    $unidata_name_pool{"<CJK ideograph>"} = -1;
	    $unidata_decomp_pool{$decomp} = -1;
	    while ($seq <= $code) {
		$unidata_name[$seq] = "<CJK ideograph>";
		$unidata_general_category[$seq] = $gencat_encoding{$gencat};
		$unidata_combining_class[$seq] = $combclass;
		$unidata_bidi_category[$seq] = $bidicat_encoding{$bidicat};
		$unidata_decomp_type[$seq] = $decomp_encoding{$decomp_type};
		$unidata_decomp[$seq] = $decomp;
		$unidata_decimal_value[$seq] = $decimal;
		$unidata_digit_value[$seq] = $digit;
		$unidata_numeric_value[2*$seq] = $numer;
		$unidata_numeric_value[2*$seq+1] = $denom;
		$unidata_upper_case[$seq] = $seq;
		$unidata_lower_case[$seq] = $seq;
		$unidata_title_case[$seq] = $seq;
		update_progress_bar ++$seq;
	    }
	}
	elsif ($name =~ /^<Hangul Syllable/) {
	    $unidata_name_pool{"<Hangul syllable>"} = -1;
	    $unidata_decomp_pool{$decomp} = -1;
	    while ($seq <= $code) {
		$unidata_name[$seq] = "<Hangul syllable>";
		$unidata_general_category[$seq] = $gencat_encoding{$gencat};
		$unidata_combining_class[$seq] = $combclass;
		$unidata_bidi_category[$seq] = $bidicat_encoding{$bidicat};
		$unidata_decomp_type[$seq] = $decomp_encoding{$decomp_type};
		$unidata_decomp[$seq] = $decomp;
		$unidata_decimal_value[$seq] = $decimal;
		$unidata_digit_value[$seq] = $digit;
		$unidata_numeric_value[2*$seq] = $numer;
		$unidata_numeric_value[2*$seq+1] = $denom;
		$unidata_upper_case[$seq] = $seq;
		$unidata_lower_case[$seq] = $seq;
		$unidata_titl_casee[$seq] = $seq;
		update_progress_bar ++$seq;
	    }
	}
	elsif ($name =~ /Surrogate, Last>$/) {
	    $unidata_name_pool{"<surrogate>"} = -1;
	    $unidata_decomp_pool{$decomp} = -1;
	    while ($seq <= $code) {
		$unidata_name[$seq] = "<surrogate>";
		$unidata_general_category[$seq] = $gencat_encoding{$gencat};
		$unidata_combining_class[$seq] = $combclass;
		$unidata_bidi_category[$seq] = $bidicat_encoding{$bidicat};
		$unidata_decomp_type[$seq] = $decomp_encoding{$decomp_type};
		$unidata_decomp[$seq] = $decomp;
		$unidata_decimal_value[$seq] = $decimal;
		$unidata_digit_value[$seq] = $digit;
		$unidata_numeric_value[2*$seq] = $numer;
		$unidata_numeric_value[2*$seq+1] = $denom;
		$unidata_upper_case[$seq] = $seq;
		$unidata_lower_case[$seq] = $seq;
		$unidata_title_case[$seq] = $seq;
		update_progress_bar ++$seq;
	    }
	}
	elsif ($name =~ /^<Private Use/) {
	    $unidata_name_pool{"<private>"} = -1;
	    $unidata_decomp_pool{$decomp} = -1;
	    while ($seq <= $code) {
		$unidata_name[$seq] = "<private>";
		$unidata_general_category[$seq] = $gencat_encoding{$gencat};
		$unidata_combining_class[$seq] = $combclass;
		$unidata_bidi_category[$seq] = $bidicat_encoding{$bidicat};
		$unidata_decomp_type[$seq] = $decomp_encoding{$decomp_type};
		$unidata_decomp[$seq] = $decomp;
		$unidata_decimal_value[$seq] = $decimal;
		$unidata_digit_value[$seq] = $digit;
		$unidata_numeric_value[2*$seq] = $numer;
		$unidata_numeric_value[2*$seq+1] = $denom;
		$unidata_upper_case[$seq] = $seq;
		$unidata_lower_case[$seq] = $seq;
		$unidata_title_case[$seq] = $seq;
		update_progress_bar ++$seq;
	    }
	}
	else {
	    panic ("Error on line $lineno of \"$src.txt\":",
		   "Unknown code range.");
	}
	$in_range = 0;
    }
    else {
	# Handle a single character:
	!$in_range or panic ("Error on line $lineno of \"$src.txt\":",
			     "Unterminated code range.");
	$unidata_name_pool{$name} = -1;
	$unidata_decomp_pool{$decomp} = -1;
	$unidata_name[$seq] = $name;
	$unidata_general_category[$seq] = $gencat_encoding{$gencat};
	$unidata_combining_class[$seq] = $combclass;
	$unidata_bidi_category[$seq] = $bidicat_encoding{$bidicat};
	$unidata_decomp_type[$seq] = $decomp_encoding{$decomp_type};
	$unidata_decomp[$seq] = $decomp;
	$unidata_decimal_value[$seq] = $decimal;
	$unidata_digit_value[$seq] = $digit;
	$unidata_numeric_value[2*$seq] = $numer;
	$unidata_numeric_value[2*$seq+1] = $denom;
	$unidata_upper_case[$seq] = $upper;
	$unidata_lower_case[$seq] = $lower;
	$unidata_title_case[$seq] = $title;
	update_progress_bar (++$seq);
    }
}
!$in_range or panic ("Error on line $lineno of \"$src.txt\":",
		     "Unterminated code range.");
while ($seq < 0xFFFE) {
    $unidata_name_pool{"<reserved>"} = -1;
    $unidata_decomp_pool{""} = -1;
    $unidata_name[$seq] = "<reserved>";
    $unidata_general_category[$seq] = $gencat_encoding{"Cn"};
    $unidata_combining_class[$seq] = 0;
    $unidata_bidi_category[$seq] = $bidicat_encoding{"L"};
    $unidata_decomp_type[$seq] = $decomp_encoding{"none"};
    $unidata_decomp[$seq] = "";
    $unidata_decimal_value[$seq] = 255;
    $unidata_digit_value[$seq] = 255;
    $unidata_numeric_value[2*$seq] = 0;
    $unidata_numeric_value[2*$seq+1] = 0;
    $unidata_upper_case[$seq] = $seq;
    $unidata_lower_case[$seq] = $seq;
    $unidata_title_case[$seq] = $seq;
    update_progress_bar ++$seq;
}
while ($seq <= 0xFFFF) {
    $unidata_name_pool{"<not a character>"} = -1;
    $unidata_decomp_pool{""} = -1;
    $unidata_name[$seq] = "<not a character>";
    $unidata_general_category[$seq] = $gencat_encoding{"Cn"};
    $unidata_combining_class[$seq] = 0;
    $unidata_bidi_category[$seq] = $bidicat_encoding{"L"};
    $unidata_decomp_type[$seq] = $decomp_encoding{"none"};
    $unidata_decomp[$seq] = "";
    $unidata_decimal_value[$seq] = 255;
    $unidata_digit_value[$seq] = 255;
    $unidata_numeric_value[2*$seq] = 0;
    $unidata_numeric_value[2*$seq+1] = 0;
    $unidata_upper_case[$seq] = $seq;
    $unidata_lower_case[$seq] = $seq;
    $unidata_title_case[$seq] = $seq;
    update_progress_bar ++$seq;
}
end_progress_bar;

# Generate "UnicodeData.c":
print STDERR "Generating \"$src.c\":\n";
open OUT, "> $src.c" or panic "Cannot create \"$src.c\"", $!;
print OUT "/* This file has been automatically generated. Do not edit! */

#include \"unidata.h\"

";

# Generate the string pool:
$table = "name_pool";
@list = sort keys %unidata_name_pool;
begin_progress_bar scalar(@list), "  Generating \"__unidata_$table\"";
print OUT "const __unidata_iso646_t __unidata_${table}[] =";
$offset = 0;
$i = 0;
foreach $string (@list) {
    $unidata_name_pool{$string} = $offset;
    print OUT "\n    \"$string\\0\"";
    update_progress_bar (++$i);
    $offset += length($string) + 1;
}
print OUT ";

";
end_progress_bar;
$name_pool_size = $offset;

# Generate the decomposition pool:
$table = "decomp_pool";
@list = sort keys %unidata_decomp_pool;
begin_progress_bar scalar(@list), "  Generating \"__unidata_$table\"";
print OUT "const __unidata_utf16_t __unidata_${table}[] = {\n";
$offset = 0;
$i = 0;
foreach $string (@list) {
    $unidata_decomp_pool{$string} = $offset;
    print OUT "    ${string}0xFFFF,\n";
    update_progress_bar (++$i);
    $offset += (length($string) + 1) * 2;
}
print OUT "};

";
end_progress_bar;
$decomp_pool_size = $offset;

# Generate the "__unidata_name" table:
$table = "name";
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_loff_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 8 == 0;
    printf OUT "%6d, ", $unidata_name_pool{$unidata_name[$i]};
    update_progress_bar $i;
}
print OUT "
};

";
end_progress_bar;

# Generate the "__unidata_general_category" table:
$table = "general_category";
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_prop_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 16 == 0;
    printf OUT "%2d, ", $unidata_general_category[$i];
    update_progress_bar $i;
}
print OUT "
};

";
end_progress_bar;

# Generate the "__unidata_combining_class" table:
$table = "combining_class";
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_prop_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 8 == 0;
    printf OUT "%3d, ", $unidata_combining_class[$i];
    update_progress_bar $i;
}
print OUT "
};

";
end_progress_bar;

# Generate the "__unidata_bidi_category" table:
$table = "bidi_category";
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_prop_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 16 == 0;
    printf OUT "%2d, ", $unidata_bidi_category[$i];
    update_progress_bar $i;
}
print OUT "

};
";
end_progress_bar;

# Generate the "__unidata_decomp_type" table:
$table = "decomp_type";
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_prop_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 16 == 0;
    printf OUT "%2d, ", $unidata_decomp_type[$i];
    update_progress_bar $i;
}
print OUT "
};

";
end_progress_bar;

# Generate the "__unidata_decomp" table:
$table = "decomp";
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_loff_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 8 == 0;
    printf OUT "%6d, ", $unidata_decomp_pool{$unidata_decomp[$i]};
    update_progress_bar $i;
}
print OUT "
};

";
end_progress_bar;

# Generate the "__unidata_decimal_value" table:
$table = "decimal_value";
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_digit_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 9 == 0;
    printf OUT "%3d, ", $unidata_decimal_value[$i];
    update_progress_bar $i;
}
print OUT "

};
";
end_progress_bar;

# Generate the "__unidata_digit_value" table:
$table = "digit_value";
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_digit_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 8 == 0;
    printf OUT "%3d, ", $unidata_digit_value[$i];
    update_progress_bar $i;
}
print OUT "

};
";
end_progress_bar;

# Generate the "__unidata_numeric_value" table:
$table = "numeric_value";
begin_progress_bar 65536*2, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_nat_t __unidata_${table}[65536*2] = {";
for ($i = 0; $i < 65536*2; ++$i) {
    print OUT "\n    " if $i % 8 == 0;
    printf OUT "%6d, ", $unidata_numeric_value[$i];
    update_progress_bar $i;
}
print OUT "

};
";
end_progress_bar;

# Generate the "__unidata_upper_case" table:
$table = "upper_case";
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_utf16_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 8 == 0;
    printf OUT "0x%04X, ", $unidata_upper_case[$i];
    update_progress_bar $i;
}
print OUT "

};
";
end_progress_bar;

# Generate the "__unidata_lower_case" table:
$table = "lower_case";
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_utf16_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 8 == 0;
    printf OUT "0x%04X, ", $unidata_lower_case[$i];
    update_progress_bar $i;
}
print OUT "

};
";
end_progress_bar;

# Generate the "__unidata_title_case" table:
$table = "title_case";
begin_progress_bar 65536, "  Generating \"__unidata_$table\"";
print OUT "const __unidata_utf16_t __unidata_${table}[65536] = {";
for ($i = 0; $i < 65536; ++$i) {
    print OUT "\n    " if $i % 8 == 0;
    printf OUT "0x%04X, ", $unidata_title_case[$i];
    update_progress_bar $i;
}
print OUT "

};
";
end_progress_bar;


print STDERR "  String pool size: $name_pool_size bytes\n";
print STDERR "  Decomposition pool size: $decomp_pool_size bytes\n";
close(OUT);
