/**************************************************************** -*- C -*- **
  2000-2001 Patryk Zadarnowski <pat@jantar.org>

 Redistribution  and  use  in  source   and  binary  forms,  with  or  without
 modification, are  permitted provided that the following  conditions are met:
 (1) Redistributions of  source code must  retain the above  copyright notice,
 this list of conditions and the following disclaimer.  (2) Redistributions in
 binary  form  must  reproduce  the  above  copyright  notice,  this  list  of
 conditions  and the following  disclaimer in  the documentation  and/or other
 materials provided with the distribution.  (3) The name of any author may not
 be used  to endorse  or promote products  derived from this  software without
 their specific prior written permission.

 THIS SOFTWARE  IS PROVIDED ``AS IS''  AND ANY EXPRESS  OR IMPLIED WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS  FOR A PARTICULAR  PURPOSE, ARE  DISCLAIMED.  IN  NO EVENT  SHALL THE
 AUTHORS BE  LIABLE FOR ANY DIRECT, INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY,
 OR  CONSEQUENTIAL DAMAGES  (INCLUDING,  BUT NOT  LIMITED  TO, PROCUREMENT  OF
 SUBSTITUTE  GOODS OR  SERVICES; LOSS  OF USE,  DATA OR  PROFITS;  OR BUSINESS
 INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY  THEORY OF  LIABILITY, WHETHER  IN
 CONTRACT,  STRICT  LIABILITY, OR  TORT  (INCLUDING  NEGLIGENCE OR  OTHERWISE)
 ARISING IN ANY  WAY OUT OF THE USE  OF THIS SOFTWARE, EVEN IF  ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.
*****************************************************************************/

/*

	The Unicode Character Database Library
	--------------------------------------------------------------

	Unicode is  a very large standard that  defined many megabytes
	of  character  information.  Leaving  this  information to  be
	encoded  privately by  every application  on demand  poses two
	problems:

	1. Unicode is  an evolving standard,  while statically-encoded
	   and linked data is difficult to update and maintain.

	2. The sheer volume of data causes excessive and pointless
	   wastage of disk space.

	"libunidata"  is   a  shared   object  encoding  all   of  the
	information  from  the  following Unicode  Character  Database
	files:

	o ArabicShaping.txt
	o BidiMirroring.txt
	o Blocks.txt
	o CaseFolding.txt
	o CompositionExclusions.txt
	o EastAsianWidth.txt
	o Jamo.txt
	o LineBreak.txt
	o UnicodeData.txt

	The following  database files  are currently not  used, either
	because they contain redundant information that can be readily
	derived from  the above files,  or because the  information is
	difficult to encode in a reasonably-general way:

	o Index.txt
	o NamesList.txt
	o NormalizationTest.txt
	o PropList.txt
	o SpecialCasing.txt
	o Unihan.txt
	o diffXvY.txt

	In a typical system, only a fraction of the database will find
	its  way into the  memory. Because  it is  loaded as  a shared
	object, the  library utilizes  the existing OS  facilities for
	sharing read-only  data and  therefore avoids wastage  of both
	RAM and disc space.

	Note that  only the basic multilingual plane  (BMP) is handled
	by the library.  The standard UTF-16 encoding already requires
	special handling of  the remaining, highly-specialized, planes
	and the  properties of those characters are  uniform enough to
	be  computed algorithmically.   Note that,  currently, neither
	Unicode  nor ISO 10646  define any  characters outside  of the
	BMP,  although inclusion  of many  specialised scripts  in the
	subsequent planes is already  being planned, and many of those
	have been accepted by the Unicode Consorium in late 2000.

	Character properties are  encoded in uncompressed tables. Each
	property is stored in a  separate table to simplify access and
	to avoid future compatibility problems.

	The  method  for encoding  string  properties and  one-to-many
	mappings  is worth  noting.   The impulse  to  encode them  as
	pointers  to arrays is  heavily misguided,  as this  would add
	several thousands  entries to the library's  symbol table, and
	dynamic linkers are simply not designed to cope with that kind
	of   load.   Therefore,   all  pointers   musto   be  resolved
	explicitely,  and   the  corresponding  database   tables  map
	characters to  offsets into  one of the  string tables  in the
	database.  The string  tables  consist of  sentinel-terminated
	lists.  Narrow  character data  is encoded using  the portable
	ISO 646 character set,  and terminated with the customary NUL.
	Wide  character data  is encoded  using UTF-16  and terminated
	with the Unicode sentinel character, U+FFFF.

	Because the library is  a database, the implementation details
	are included  in the specification. The numeric  values of the
	constants  defined below are  normative and  you can  count on
	them not to be changed in the future revisions of the library,
	although new values and tables  may be added as necessary. All
	external   symbols,  types   and  constants   are   given  the
	"__unidata"  or "_Unidata"  prefix to  avoid pollution  of the
	user name space.  The  library is intended for transparent use
	by other libraries, and  therefore cannot occupy a public name
	space.

*/

#ifndef __unidata_h_included
#define __unidata_h_included

#include <limits.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/*

	Versions
	--------------------------------------------------------------

	Unicode  versions consist  of three  components,  described in
	2.10 of The Unicode Standard, Version 3.0.

*/

extern const int __unidata_version_major;
extern const int __unidata_version_minor;
extern const int __unidata_version_update;

/*

	Types
	--------------------------------------------------------------

	Before we can declare the  mapping tables, we must define some
	types. Note that  I do not use wchar_t, as it  is likely to be
	compiler-dependant,  and  the  library  must  remain  portable
	within an architecture.

	The  two types  "__unidata_soff_t" and  "__unidata_loff_t" are
	used to  represent byte  offsets. The first  is used  when the
	string pool  is known to  be small, and therefore  the offsets
	may be represented  in 16 bits.  Otherwise, we  use the 32 bit
	type "__unidata_loff_t" to index the string pool.

	The "__unidata_digit_t" and  "__unidata_nat_t" types deserve a
	special mention. The  first is used to store  digit values and
	is a simple 8 bit unsigned integer type (currently, all digits
	recognized  by Unicode are  decimal.)  The  second is  used to
	represent natural  numbers, which, in turn, are  used to store
	the numeric value property. This property maps some characters
	to  rational numbers  p/q, which  are  encoded as  a pair  (p,
	q).  Currently, the  largest natural  number appearing  in the
	standard  is 10000,  and this  is  unlikely to  change in  the
	future, so  I use unsigned 16  bit integer type  to encode the
	numeric value.

	Note that,  below, "short" and  "int" are used  unportably as,
	the  first is  assumed to  be exactly  16 bits  wide,  and the
	second is assumed to be at  least 32 bits wide.  I verify this
	requirement using constants from <limits.h>.

*/

#if SHRT_MAX != 0x7fff
# error "short" is not a 16 bit type!
#endif

#if INT_MAX < 0x7FFFFFFF
# error "int" is not a 32 bit type!
#endif

typedef unsigned char  __unidata_iso646_t;
typedef unsigned short __unidata_utf16_t;
typedef unsigned char  __unidata_prop_t;
typedef unsigned int   __unidata_flags_t;
typedef unsigned short __unidata_soff_t;
typedef unsigned int   __unidata_loff_t;
typedef unsigned char  __unidata_digit_t;
typedef unsigned short __unidata_nat_t;

/*

	Arabic Shaping
	--------------------------------------------------------------

	Arabic    shaping   properties    are    derived   from    the
	"ArabicShaping.txt" database file,  and described in section 8
	of  The Unicode  Standard, Version  3.  Currently,  the source
	database provided  by Unicode  is very incomplete,  and should
	probably be complemented manually  by the information from the
	Unicode book.  I may  do this in  some future revision  of the
	library.

*/

#define	_Unidata_arabic_R	0 /* right_joining */
#define	_Unidata_arabic_L	1 /* left_joining  */
#define	_Unidata_arabic_D	2 /* dual_joining  */
#define	_Unidata_arabic_C	3 /* join_causing  */
#define	_Unidata_arabic_N	4 /* non_joining   */
#define	_Unidata_arabic_T	5 /* transparent   */

extern const __unidata_iso646_t __unidata_arabic_shaping_pool[];
extern const __unidata_soff_t   __unidata_arabic_schematic_name[65536];
extern const __unidata_prop_t   __unidata_arabic_joining_type[65536];
extern const __unidata_soff_t   __unidata_arabic_joining_group[65536];

/*

	Bidirectional Mirroring
	--------------------------------------------------------------

	This  information  is  derived  from  the  "BidiMirroring.txt"
	database file introduced in version 3.0.1 of the standard, and
	describes  the  bidirectional  mirroring property  of  Unicode
	characters. The  semantics of  this property are  described in
	UTR #21: Bidirectional Algorithm,

*/

#define _Unidata_unmirrored      0 /* char. without the "mirrored" property */
#define _Unidata_no_mirror       1 /* no mirror mapping within the BMP */
#define _Unidata_best_fit_mirror 2 /* only a best-fit mapping avaliable */
#define _Unidata_proper_mirror   3 /* a proper mirror image avaliable */

extern const __unidata_prop_t  __unidata_bidi_mapping_type[65536];
extern const __unidata_utf16_t __unidata_bidi_mirror[65536];

/*

	Character block names
	--------------------------------------------------------------

	This data is derived  from the "Blocks.txt" database file, and
	describes the  layout of  the basic multilingual  plane.  This
	information is  structured in a somewhat  unusual fashion. The
	BMP  is  divided   into  non-overlapping  blocks,  with  bogus
	``reserved'' blocks inserted to ensure a complete cover by the
	partitioning.   The number of  blocks, including  the reserved
	ones,  is stored in  "__unidata_block_count".  The  first code
	position  of each block  is stored  in "__unidata_block_posn".
	To make it  possible to compute the block  size, an additional
	U+0000  entry  has  been appended  to  "__unidata_block_posn",
	which can be used both as a list terminator and to compute the
	block size  as a difference of the  two consecutive positions,
	modulo  65536.    The  name  of   each  block  is   stored  in
	"__unidata_block_name",  as  a  byte  offset into  the  string
	pool.  Unallocated blocks  are  given names  consisting of  an
	empty string.

*/

extern const int __unidata_block_count;
extern const __unidata_iso646_t __unidata_block_pool[];
extern const __unidata_utf16_t  __unidata_block_posn[];
extern const __unidata_soff_t   __unidata_block_name[];

/*

	Case folding
	--------------------------------------------------------------

	This data is derived from the "CaseFolding.txt" database file,
	introduced  in version 3.0.1  of the  standard. It  defines an
	algorithm for eliminating  case information from Unicode text.
	Because  not all  letters  have proper  lower- and  upper-case
	varieties, the  tranformation is not reversible,  and does not
	produce a  proper lower-case  version of the  string. Instead,
	the result  of this  mapping is indented  for case-insensitive
	text  searching  and similar  applications.  The algorithm  is
	covered in detail in UTR #21.

*/

#define _Unidata_no_case_mapping		0
#define _Unidata_lower_case_mapping		1
#define _Unidata_exceptional_case_mapping	2

extern const __unidata_utf16_t __unidata_neutral_case_pool[];
extern const __unidata_prop_t  __unidata_neutral_case_mapping_type[65536];
extern const __unidata_loff_t  __unidata_neutral_case[65536];

/*

	Composition exclusions
	--------------------------------------------------------------

	This elusive but important  data describes the characters from
	the UTR  #15 Composition Exclusion  Table. It is  derived from
	the "CompositionExclusions.txt" and "UnicodeData.txt" files as
	described in the comment in "CompositionExclusions.txt".

*/

#define _Unidata_no_exclusion     0 /* char. not exluded from decomposition */
#define	_Unidata_script_exclusion 1 /* a script-specific exclusion          */
#define	_Unidata_post_exclusion   2 /* post-composition version character   */
#define	_Unidata_singleton        3 /* singleton decomposition              */
#define	_Unidata_non_starter      4 /* non-starter decomposition            */

extern const __unidata_prop_t __unidata_composition_exclusion[65536];

/*

	East-Asian character width
	--------------------------------------------------------------

	This  data is derived  from the  "EastAsianWidth.txt" database
	file,   and  describes  the   East-Asian  width   property  of
	characters.

*/

#define _Unidata_east_asian_neutral 	0 /* East Asian neutral    */
#define _Unidata_east_asian_ambiguous	1 /* East Asian ambiguous  */
#define _Unidata_east_asian_half_width	2 /* East Asian half-width */
#define _Unidata_east_asian_wide	3 /* East Asian wide       */
#define _Unidata_east_asian_full_width	4 /* East Asian full-width */
#define _Unidata_east_asian_narrow	5 /* East Asian narrow     */

extern const __unidata_prop_t   __unidata_east_asian_width[65536];

/*

	Jamo short names
	--------------------------------------------------------------

	This data  is derived from  the "Jamo.txt" database  file, and
	describes the Jamo Short Name property of Unicode characters.

	This property, used to compute the names of Hangul characters,
	is only applicable to a small region of the BMP, and the below
	tables  map   out  only  the  code  positions   in  the  range
	[U+1100, U+11FF].

*/

#define __Unidata_jamo_offset 0x1100

extern const __unidata_iso646_t __unidata_jamo_pool[];
extern const __unidata_soff_t   __unidata_jamo_short_name[256];

/*

	Line break properties
	--------------------------------------------------------------

	This data  is derived from the  "LineBreak.txt" database file,
	and  describes   the  line  breaking   properties  of  Unicode
	characters.

*/

#define _Unidata_XX  0 /* unknown */
#define _Unidata_BK  1 /* mandatory break */
#define _Unidata_CR  2 /* carriage return */
#define _Unidata_LF  3 /* line feed */
#define _Unidata_CM  4 /* attached characters and combining marks marks */
#define _Unidata_SG  5 /* surrogates */
#define _Unidata_ZW  6 /* zero width space */
#define _Unidata_IN  7 /* inseparable */
#define _Unidata_GL  8 /* non-breaking (glue) */
#define _Unidata_CB  9 /* contingent break opportunity */
#define _Unidata_SP 10 /* space */
#define _Unidata_BA 11 /* break opportunity after */
#define _Unidata_BB 12 /* break opportunity before */
#define _Unidata_B2 13 /* break opportunity before and after */
#define _Unidata_HY 14 /* hyphen */
#define _Unidata_NS 15 /* non starter */
#define _Unidata_OP 16 /* opening punctuation */
#define _Unidata_CL 17 /* closing punctuation */
#define _Unidata_QU 18 /* ambiguous quotation */
#define _Unidata_EX 19 /* exclamation/interrogation */
#define _Unidata_ID 20 /* ideographic */
#define _Unidata_NU 21 /* numeric */
#define _Unidata_IS 22 /* infix separator (numeric) */
#define _Unidata_SY 23 /* symbols allowing breaks */
#define _Unidata_AL 24 /* ordinary alphabetic and symbol characters */
#define _Unidata_PR 25 /* prefix (numeric) */
#define _Unidata_PO 26 /* postfix (numeric) */
#define _Unidata_SA 27 /* complex context (south east asian) */
#define _Unidata_AI 28 /* ambiguous (alphabetic or ideographic) */

extern const __unidata_prop_t  __unidata_line_break_type[65536];

/*

	Character names
	--------------------------------------------------------------

	This data has relatively few uses, but is an important part of
	the standard. It may be used, for example, to display the name
	of  the character  at the  cursor in  a status  bar of  a text
	editor.

*/

extern const __unidata_iso646_t __unidata_name_pool[];
extern const __unidata_loff_t   __unidata_name[65536];

/*

	General category
	--------------------------------------------------------------

	This data provides access to the general category of each code
	position.   Note that Unicode  gurantees that  every character
	belongs to exactly one  category. This simplifies the encoding
	of  the property:  rather  than the  usual  bit mask,  general
	category is stored  as en enumerated property. One  can test a
	character for a set of categories simultaneously by generating
	the  appropriate  bit   set  dynamically.   For  example,  the
	following expression implements "iswalpha" by checkiing if "x"
	is a letter:

		((1 << __unidata_general_category[x]) &
		 ((1 << __unidata_Lu) |
		  (1 << __unidata_Ll) |
		  (1 << __unidata_Lt) |
		  (1 << __unidata_Lm) |
		  (1 << __unidata_Lo)))

	with  only  minimal  impact   on  the  amount  of  computation
	performed  by the CPU,  and a  substential improvement  of the
	data cache usage pattern.

*/

#define _Unidata_Lu	 0 /* Letter, Uppercase          */
#define _Unidata_Ll	 1 /* Letter, Lowercase          */
#define _Unidata_Lt	 2 /* Letter, Titlecase          */
#define _Unidata_Mn	 3 /* Mark, Non-Spacing          */
#define _Unidata_Mc	 4 /* Mark, Spacing Combining    */
#define _Unidata_Me	 5 /* Mark, Enclosing            */
#define _Unidata_Nd	 6 /* Number, Decimal Digit      */
#define _Unidata_Nl	 7 /* Number, Letter             */
#define _Unidata_No	 8 /* Number, Other              */
#define _Unidata_Zs	 9 /* Separator, Space           */
#define _Unidata_Zl	10 /* Separator, Line            */
#define _Unidata_Zp	11 /* Separator, Paragraph       */
#define _Unidata_Cc	12 /* Other, Control             */
#define _Unidata_Cf	13 /* Other, Format              */
#define _Unidata_Cs	14 /* Other, Surrogate           */
#define _Unidata_Co	15 /* Other, Private Use         */
#define _Unidata_Cn	16 /* Other, Not Assigned        */
#define _Unidata_Lm	17 /* Letter, Modifier           */
#define _Unidata_Lo	18 /* Letter, Other              */
#define _Unidata_Pc	19 /* Punctuation, Connector     */
#define _Unidata_Pd	20 /* Punctuation, Dash          */
#define _Unidata_Ps	21 /* Punctuation, Open          */
#define _Unidata_Pe	22 /* Punctuation, Close         */
#define _Unidata_Pi	23 /* Punctuation, Initial quote */
#define _Unidata_Pf	24 /* Punctuation, Final quote   */
#define _Unidata_Po	25 /* Punctuation, Other         */
#define _Unidata_Sm	26 /* Symbol, Math               */
#define _Unidata_Sc	27 /* Symbol, Currency           */
#define _Unidata_Sk	28 /* Symbol, Modifier           */
#define _Unidata_So	29 /* Symbol, Other              */

extern const __unidata_prop_t __unidata_general_category[65536];

/*

	Canonical Combining Classes
	--------------------------------------------------------------

	Canonical combining classes are assigned numeric codes between
	0 and  240 inclusively,  and are stored  as such in  the below
	table.

*/

extern const __unidata_prop_t __unidata_combining_class[65536];

/*

	Bidirectional Category
	--------------------------------------------------------------

	Bidirectional category is a simple enumerated property of each
	character, encoded as specified below.

*/

#define _Unidata_bidi_L		 0 /* Left-to-Right              */
#define _Unidata_bidi_LRE	 1 /* Left-to-Right Embedding    */
#define _Unidata_bidi_LRO	 2 /* Left-to-Right Override     */
#define _Unidata_bidi_R		 3 /* Right-to-Left              */
#define _Unidata_bidi_AL	 4 /* Right-to-Left Arabic       */
#define _Unidata_bidi_RLE	 5 /* Right-to-Left Embedding    */
#define _Unidata_bidi_RLO	 6 /* Right-to-Left Override     */
#define _Unidata_bidi_PDF	 7 /* Pop Directional Format     */
#define _Unidata_bidi_EN	 8 /* European Number            */
#define _Unidata_bidi_ES	 9 /* European Number Separator  */
#define _Unidata_bidi_ET	10 /* European Number Terminator */
#define _Unidata_bidi_AN	11 /* Arabic Number              */
#define _Unidata_bidi_CS	12 /* Common Number Separator    */
#define _Unidata_bidi_NSM	13 /* Non-Spacing Mark           */
#define _Unidata_bidi_BN	14 /* Boundary Neutral           */
#define _Unidata_bidi_B		15 /* Paragraph Separator        */
#define _Unidata_bidi_S		16 /* Segment Separator          */
#define _Unidata_bidi_WS	17 /* Whitespace                 */
#define _Unidata_bidi_ON	18 /* Other Neutrals             */

extern const __unidata_prop_t __unidata_bidi_category[65536];

/*

	Character Decomposition Mapping
	--------------------------------------------------------------

	The decomposition  mapping is an  important property. It  is a
	one-to-many mapping,  and is therefore stored  using the usual
	string  pool technique.   Characters  without a  decomposition
	mapping are mapped to an empty string, located at the offset 0
	in the decomposition string pool.

*/

#define __unidata_none      0 /* No decomposition mapping                    */
#define __unidata_canon     1 /* A canononical decomposition mapping         */
#define __unidata_font      2 /* A font variant (eg. a blackletter form)     */
#define __unidata_no_break  3 /* A no-break version of a space or hyphen     */
#define __unidata_initial   4 /* An initial presentation form (Arabic)       */
#define __unidata_medial    5 /* A medial presentation form (Arabic)         */
#define __unidata_final     6 /* A final presentation form (Arabic)          */
#define __unidata_isolated  7 /* An isolated presentation form (Arabic)      */
#define __unidata_circle    8 /* An encircled form                           */
#define __unidata_super	    9 /* A superscript form                          */
#define __unidata_sub      10 /* A subscript form                            */
#define __unidata_vertical 11 /* A vertical layout presentation form         */
#define __unidata_wide     12 /* A wide (or zenkaku) compatibility character */
#define __unidata_narrow   13 /* A narrow (or hankaku) compatibility char.   */
#define __unidata_small    14 /* A small variant form (CNS compatibility)    */
#define __unidata_square   15 /* A CJK squared font variant                  */
#define __unidata_fraction 16 /* A vulgar fraction form                      */
#define __unidata_compat   17 /* Otherwise unspecified compatibility char.   */

extern const __unidata_utf16_t __unidata_decomp_pool[];
extern const __unidata_prop_t  __unidata_decomp_type[65536];
extern const __unidata_loff_t  __unidata_decomp[65536];

/*

	Decimal Digit, Digit and Numeric Values
	--------------------------------------------------------------

	These three  tables map out the numeric  properties of Unicode
	characters.   Non-digits map  to the  value  255.  Non-numeric
	characters map to the numeric value of 0/0.

*/

extern const __unidata_digit_t __unidata_decimal_value[65536];
extern const __unidata_digit_t __unidata_digit_value[65536];
extern const __unidata_nat_t   __unidata_numeric_value[65536*2];

/*

	Upper-, Lower- And Title-Case Mappings
	--------------------------------------------------------------

	These three  tables store one-to-one case  mappings. Note that
	these properties, although  normative, are somewhat deceptive:
	the information from "SpecialCasing.txt" is required to obtain
	a  complete  (possibly  one-to-many,  many-to-one  or  locale-
	specific) case mapping, and  this information is currently not
	provided by  the library.  All characters that  do not  have a
	mapping defined in "UnicodeData.txt" are mapped to themselves.

*/

extern const __unidata_utf16_t __unidata_upper_case[65536];
extern const __unidata_utf16_t __unidata_lower_case[65536];
extern const __unidata_utf16_t __unidata_title_case[65536];

/*

	Block Names
	--------------------------------------------------------------

	The "__unidata_block_posn" array is a list of the positions of
	each  block in  the  16-bit Unicode  code  space. It  contains
	"__unidata_block_count+1"    entries.     The    blocks    are
	non-overlapping,   and  the   list  is   made   continuous  by
	introducing   ``undefined''   blocks   with  null   names   as
	necessary.  We store  0 in  an additional  final entry  in the
	list,  which  can  be  used  both as  a  terminator  and  when
	computing block sizes, using the fact that 0 == 0x10000 modulo
	2**16.

	The  "__unicode_block_name"  array stores  the  names of  each
	block   as  byte   offsets  into   the  "__unicode_block_pool"
	table. It  contains "__unidata_block_count" entries,  ie. does
	not index the additional null entry in "__unidata_block_posn".

	Finally, the  "__unicode_block_pool" array stores  a free list
	of  NUL-terminated  ISO  646  strings representing  the  block
	names.

*/

extern const __unidata_utf16_t  __unidata_block_posn[];
extern const __unidata_soff_t   __unidata_block_name[];
extern const __unidata_iso646_t __unidata_block_pool[];
extern const int                __unidata_block_count;

/*

	Jamo Short Names
	--------------------------------------------------------------

	These two tables  map out the Jamo short  name property of the
	Hangul Jamo block.

*/

extern const __unidata_soff_t   __unidata_jamo[256];
extern const __unidata_iso646_t __unidata_jamo_pool[];

/*

	Arabic Shaping
	--------------------------------------------------------------

*/

extern const __unidata_prop_t   __unidata_arabic_joining_type[65536];
extern const __unidata_soff_t   __unidata_arabic_joining_group[65536];
extern const __unidata_iso646_t __unidata_arabic_joining_pool[];

/*

	Case Folding
	--------------------------------------------------------------

*/

extern const __unidata_loff_t  __unidata_case[65536];
extern const __unidata_utf16_t __unidata_case_pool[];

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __unidata_h_included */
